import os
import re
from pathlib import Path
from typing import Any, Iterable, Sequence, Union, List, Tuple, Dict

from docutils import nodes
from docutils.nodes import Node, document

from sphinx.application import Sphinx
from sphinx.application import logger as sphinx_log
try:
    from sphinx.builders.dirhtml import DirectoryHTMLBuilder
except ImportError:
    # Try importing from older Sphinx version path.
    from sphinx.builders.html import DirectoryHTMLBuilder
from sphinx.config import Config
from docutils.parsers.rst import Directive

needs_sphinx = "1.7" # Do not require newer sphinx. EPEL packages build man pages with Sphinx 1.7.6. Refer: CDRIVER-4767
author = "MongoDB, Inc"

# -- Options for HTML output ----------------------------------------------

smart_quotes = False
html_show_sourcelink = False

# Note: http://www.sphinx-doc.org/en/1.5.1/config.html#confval-html_copy_source
# This will degrade the Javascript quicksearch if we ever use it.
html_copy_source = False


def _file_man_page_name(fpath: Path) -> Union[str, None]:
    "Given an rST file input, find the :man_page: frontmatter value, if present"
    lines = fpath.read_text().splitlines()
    for line in lines:
        mat = re.match(r":man_page:\s+(.+)", line)
        if not mat:
            continue
        return mat[1]

def _collect_man (app: Sphinx):
    # Note: 'app' is partially-formed, as this is called from the Sphinx.__init__
    docdir = Path(app.srcdir)
    # Find everything:
    children = docdir.rglob("*")
    # Find only regular files:
    files = filter(Path.is_file, children)
    # Find files that have a .rst extension:
    rst_files = (f for f in files if f.suffix == ".rst")
    # Pair each file with its :man_page: frontmatter, if present:
    with_man_name = ((f, _file_man_page_name(f)) for f in rst_files)
    # Filter out pages that do not have a :man_page: item:s
    pairs: Iterable[tuple[Path, str]] = ((f, m) for f, m in with_man_name if m)
    # Populate the man_pages:
    for filepath, man_name in pairs:
        # Generate the docname.
        relative_path = filepath.relative_to(app.srcdir)
        # The docname is relative to the source directory, and excludes the suffix.
        docname = str(relative_path.parent / filepath.stem)

        assert docname, filepath
        man_pages.append((docname, man_name, "", [author], 3))

# -- Options for manual page output ---------------------------------------

# NOTE: This starts empty, but we populate it in `setup` in _collect_man() (see above)
man_pages: List[Tuple[str, str, str, List[str], int]] = []

# If true, show URL addresses after external links.
#
# man_show_urls = False

# -- Sphinx customization ---------------------------------------


def add_ga_javascript(app: Sphinx, pagename: str, templatename: str, context: Dict[str, Any], doctree: document):
    if not app.env.config.analytics:
        return

    # Add google analytics and NPS survey.
    context["metatags"] = (
        context.get("metatags", "")
        + """
<!-- Google tag (gtag.js) -->
<script async src="https://www.googletagmanager.com/gtag/js?id=G-56KD6L3MDX"></script>
<script>
  window.dataLayer = window.dataLayer || [];
  function gtag(){dataLayer.push(arguments);}
  gtag('js', new Date());

  gtag('config', 'G-56KD6L3MDX');
</script>
<!--  NPS survey -->
<script type="text/javascript">
  !function(e,t,r,n,a){if(!e[a]){for(var i=e[a]=[],s=0;s<r.length;s++){var c=r[s];i[c]=i[c]||function(e){return function(){var t=Array.prototype.slice.call(arguments);i.push([e,t])}}(c)}i.SNIPPET_VERSION="1.0.1";var o=t.createElement("script");o.type="text/javascript",o.async=!0,o.src="https://d2yyd1h5u9mauk.cloudfront.net/integrations/web/v1/library/"+n+"/"+a+".js";var l=t.getElementsByTagName("script")[0];l.parentNode.insertBefore(o,l)}}(window,document,["survey","reset","config","init","set","get","event","identify","track","page","screen","group","alias"],"Dk30CC86ba0nATlK","delighted");

  delighted.survey();
</script>
"""
    )


class VersionList(Directive):
    """Custom directive to generate the version list from an environment variable"""

    option_spec = {}
    has_content = True

    def run(self) -> Sequence[Node]:
        if self.content[0] != "libmongoc" and self.content[0] != "libbson":
            print("versionlist must be libmongoc or libbson")
            return []

        libname = self.content[0]
        env_name = libname.upper() + "_VERSION_LIST"
        if env_name not in os.environ:
            print(env_name + " not set, not generating version list")
            return []

        versions = os.environ[env_name].split(",")

        header = nodes.paragraph("", "")
        p = nodes.paragraph("", "")
        uri = "https://www.mongoc.org/%s/%s/index.html" % (libname, versions[0])
        p += nodes.reference("", "Latest Release (%s)" % versions[0], internal=False, refuri=uri)
        header += p
        p = nodes.paragraph("", "")
        uri = "https://s3.amazonaws.com/mciuploads/mongo-c-driver/docs/%s/latest/index.html" % (libname)
        p += nodes.reference("", "Current Development (master)", internal=False, refuri=uri)
        header += p

        blist = nodes.bullet_list()
        for v in versions:
            item = nodes.list_item()
            p = nodes.paragraph("", "")
            uri = "https://www.mongoc.org/%s/%s/index.html" % (libname, v)
            p += nodes.reference("", v, internal=False, refuri=uri)
            item += p
            blist += item
        return [header, blist]


def generate_html_redirs(app: Sphinx, page: str, templatename: str, context: Dict[str, Any], doctree: Any) -> None:
    builder = app.builder
    if not isinstance(builder, DirectoryHTMLBuilder) or "writing-redirect" in context:
        return
    if page == "index" or page.endswith(".index"):
        return
    path = app.project.doc2path(page, absolute=True)
    out_index_html = Path(builder.get_outfilename(page))
    slug = out_index_html.parent.name
    redirect_file = out_index_html.parent.parent / f"{slug}.html"
    # HACK: handle_page() is not properly reentrant. Save and restore state for
    # this page while we generate our redirects page:
    prev_scripts = builder.script_files[:]
    prev_css = builder.css_files[:]
    builder.handle_page(
        f"redirect-for-{page}",
        {"target": page, "writing-redirect": 1},
        str(Path(__file__).parent.resolve() / "redirect.t.html"),
        str(redirect_file),
    )
    # Restore prior state:
    builder.script_files[:] = prev_scripts
    builder.css_files[:] = prev_css
    sphinx_log.debug("Wrote redirect: %r -> %r", path, page)

def mongoc_common_setup(app: Sphinx):
    _collect_man(app)
    app.connect("html-page-context", generate_html_redirs)
    app.connect("html-page-context", add_ga_javascript)
    # Run sphinx-build -D analytics=1 to enable Google Analytics.
    app.add_config_value("analytics", False, "html")
    app.add_directive("versionlist", VersionList)



#!/usr/bin/env python
#
# Copyright 2009-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""IDL for functions that take flexible options as a bson_t.

Defines the options accepted by functions that receive a const bson_t *opts,
for example mongoc_collection_find_with_opts, mongoc_collection_insert_one,
and many others.

Generates struct types, options parsing code, and RST documentation.

Written for Python 2.6+, requires Jinja 2 for templating.
"""

# yapf: disable
from collections import OrderedDict
from os.path import basename, dirname, join as joinpath, normpath
import re

from jinja2 import Environment, FileSystemLoader  # Please "pip install jinja2".

this_dir = dirname(__file__)
template_dir = joinpath(this_dir, 'opts_templates')
src_dir = normpath(joinpath(this_dir, '../src/libmongoc/src/mongoc'))
doc_includes = normpath(joinpath(this_dir, '../src/libmongoc/doc/includes'))


def flatten(items):
    for item in items:
        if isinstance(item, list):
            # "yield from".
            for subitem in flatten(item):
                yield subitem
        else:
            yield item


class Struct(OrderedDict):
    def __init__(self, items, opts_name='opts', generate_rst=True,
                 generate_code=True, allow_extra=True, **defaults):
        """Define an options struct.

        - items: List of pairs: (optionName, info)
        - opts_name: Name of the const bson_t *opts parameter
        - allow_extra: Whether to allow unrecognized options
        - defaults: Initial values for options
        """
        OrderedDict.__init__(self, list(flatten(items)))
        self.is_shared = False
        self.opts_name = opts_name
        self.generate_rst = generate_rst
        self.generate_code = generate_code
        self.allow_extra = allow_extra
        self.defaults = defaults

    def default(self, item, fallback):
        return self.defaults.get(item, fallback)


class Shared(Struct):
    def __init__(self, items, **defaults):
        """Define a struct that is shared by others."""
        super(Shared, self).__init__(items, **defaults)
        self.is_shared = True
        self.generate_rst = False

read_concern_help = 'Construct a :symbol:`mongoc_read_concern_t` and use :symbol:`mongoc_read_concern_append` to add the read concern to ``opts``. See the example code for :symbol:`mongoc_client_read_command_with_opts`. Read concern requires MongoDB 3.2 or later, otherwise an error is returned.'
read_concern_document_option = ('readConcern', {
    'type': 'document',
    'help': read_concern_help
})

read_concern_option = ('readConcern', {
    'type': 'mongoc_read_concern_t *',
    'help': read_concern_help,
    'convert': '_mongoc_convert_read_concern'
})

write_concern_option = [
    ('writeConcern', {
        'type': 'mongoc_write_concern_t *',
        'convert': '_mongoc_convert_write_concern',
        'help': 'Construct a :symbol:`mongoc_write_concern_t` and use :symbol:`mongoc_write_concern_append` to add the write concern to ``opts``. See the example code for :symbol:`mongoc_client_write_command_with_opts`.'
    }),
    ('write_concern_owned', {
        'type': 'bool',
        'internal': True,
    })
]

session_option = ('sessionId', {
    'type': 'mongoc_client_session_t *',
    'convert': '_mongoc_convert_session_id',
    'field': 'client_session',
    'help': 'First, construct a :symbol:`mongoc_client_session_t` with :symbol:`mongoc_client_start_session`. You can begin a transaction with :symbol:`mongoc_client_session_start_transaction`, optionally with a :symbol:`mongoc_transaction_opt_t` that overrides the options inherited from |opts-source|, and use :symbol:`mongoc_client_session_append` to add the session to ``opts``. See the example code for :symbol:`mongoc_client_session_t`.'
})

ordered_option = ('ordered', {
    'type': 'bool',
    'help': 'set to ``false`` to attempt to insert all documents, continuing after errors.'
})

validate_option = ('validate', {
    'type': 'bson_validate_flags_t',
    'convert': '_mongoc_convert_validate_flags',
    'help': 'Construct a bitwise-or of all desired :symbol:`bson_validate_flags_t`. Set to ``false`` to skip client-side validation of the provided BSON documents.'
})

collation_option = ('collation', {
    'type': 'document',
    'help': 'Configure textual comparisons. See `Setting Collation Order <setting_collation_order_>`_, and `the MongoDB Manual entry on Collation <https://www.mongodb.com/docs/manual/reference/collation/>`_. Collation requires MongoDB 3.2 or later, otherwise an error is returned.'
})

array_filters_option = ('arrayFilters', {
    'type': 'array',
    'help': 'An array of filters specifying to which array elements an update should apply.',
})

upsert_option = ('upsert', {
    'type': 'bool',
    'help': 'When true, creates a new document if no document matches the query.'
})

bypass_option = ('bypassDocumentValidation', {
    'type': 'bool',
    'field': 'bypass',
    'help': 'Set to ``true`` to skip server-side schema validation of the provided BSON documents.'
})

server_option = ('serverId', {
    'type': 'uint32_t',
    'convert': '_mongoc_convert_server_id',
    'help': 'To target a specific server, include an int32 "serverId" field. Obtain the id by calling :symbol:`mongoc_client_select_server`, then :symbol:`mongoc_server_description_id` on its return value.'
})

hint_option = ('hint', {
    'type': 'bson_value_t',
    'convert': '_mongoc_convert_hint',
    'help': 'A document or string that specifies the index to use to support the query predicate.'
})

let_option = ('let', {
    'type': 'document',
    'help': 'A BSON document consisting of any number of parameter names, each followed by definitions of constants in the MQL Aggregate Expression language.'
})

comment_option_since_4_4 = ('comment', {
    'type': 'bson_value_t',
    'convert': '_mongoc_convert_bson_value_t',
    'help': 'A :symbol:`bson_value_t` specifying the comment to attach to this command. The comment will appear in log messages, profiler output, and currentOp output. Requires MongoDB 4.4 or later.'
})

comment_option_string_pre_4_4 = ('comment', {
    'type': 'bson_value_t',
    'convert': '_mongoc_convert_bson_value_t',
    'help': 'A :symbol:`bson_value_t` specifying the comment to attach to this command. The comment will appear in log messages, profiler output, and currentOp output. Only string values are supported prior to MongoDB 4.4.'
})

opts_structs = OrderedDict([
    ('mongoc_crud_opts_t', Shared([
        write_concern_option,
        session_option,
        validate_option,
        comment_option_since_4_4,
    ])),

    ('mongoc_update_opts_t', Shared([
        ('crud', {'type': 'mongoc_crud_opts_t'}),
        bypass_option,
        collation_option,
        hint_option,
        upsert_option,
        let_option,
    ])),

    ('mongoc_insert_one_opts_t', Struct([
        ('crud', {'type': 'mongoc_crud_opts_t'}),
        bypass_option
    ], validate='_mongoc_default_insert_vflags')),

    ('mongoc_insert_many_opts_t', Struct([
        ('crud', {'type': 'mongoc_crud_opts_t'}),
        ordered_option,
        bypass_option,
    ], validate='_mongoc_default_insert_vflags', ordered='true')),

    ('mongoc_delete_opts_t', Shared([
        ('crud', {'type': 'mongoc_crud_opts_t'}),
        collation_option,
        hint_option,
        let_option,
    ])),

    ('mongoc_delete_one_opts_t', Struct([
        ('delete', {'type': 'mongoc_delete_opts_t'}),
    ])),

    ('mongoc_delete_many_opts_t', Struct([
        ('delete', {'type': 'mongoc_delete_opts_t'}),
    ])),

    ('mongoc_update_one_opts_t', Struct([
        ('update', {'type': 'mongoc_update_opts_t'}),
        array_filters_option,
        ('sort', {'type': 'document', 'help': 'Specify a sort order when matching documents.'})
    ], validate='_mongoc_default_update_vflags')),

    ('mongoc_update_many_opts_t', Struct([
        ('update', {'type': 'mongoc_update_opts_t'}),
        array_filters_option,
    ], validate='_mongoc_default_update_vflags')),

    ('mongoc_replace_one_opts_t', Struct([
        ('update', {'type': 'mongoc_update_opts_t'}),
        ('sort', {'type': 'document', 'help': 'Specify a sort order when matching documents.'})
    ], validate='_mongoc_default_replace_vflags')),

    ('mongoc_bulk_opts_t', Struct([
        write_concern_option,
        ordered_option,
        session_option,
        let_option,
        comment_option_since_4_4,
    ], allow_extra=False, ordered='true')),

    ('mongoc_bulk_insert_opts_t', Struct([
        validate_option,
    ], validate='_mongoc_default_insert_vflags', allow_extra=False)),

    ('mongoc_bulk_update_opts_t', Shared([
        validate_option,
        collation_option,
        hint_option,
        ('upsert', {
            'type': 'bool',
            'help': 'If true, insert a document if none match ``selector``.'
        }),
        ('multi', {'type': 'bool', 'hidden': True})
    ])),

    ('mongoc_bulk_update_one_opts_t', Struct(
        [
            ('update', {'type': 'mongoc_bulk_update_opts_t'}),
            ('sort', {'type': 'document', 'help': 'Specify a sort order when matching documents.'}),
            array_filters_option,
        ],
        multi='false',
        validate='_mongoc_default_update_vflags',
        allow_extra=False)),

    ('mongoc_bulk_update_many_opts_t', Struct(
        [
            ('update', {'type': 'mongoc_bulk_update_opts_t'}),
            array_filters_option,
        ],
        multi='true',
        validate='_mongoc_default_update_vflags',
        allow_extra=False)),

    ('mongoc_bulk_replace_one_opts_t', Struct(
        [
            ('update', {'type': 'mongoc_bulk_update_opts_t'}),
            ('sort', {'type': 'document', 'help': 'Specify a sort order when matching documents.'})
        ],
        multi='false',
        validate='_mongoc_default_replace_vflags',
        allow_extra=False)),

    ('mongoc_bulk_remove_opts_t', Shared([
        collation_option,
        hint_option,
        ('limit', {'type': 'int32_t', 'hidden': True})
    ])),

    ('mongoc_bulk_remove_one_opts_t', Struct([
        ('remove', {'type': 'mongoc_bulk_remove_opts_t'}),
    ], limit=1, allow_extra=False)),

    ('mongoc_bulk_remove_many_opts_t', Struct([
        ('remove', {'type': 'mongoc_bulk_remove_opts_t'}),
    ], limit=0, allow_extra=False)),

    ('mongoc_change_stream_opts_t', Struct([
        ('batchSize', {'type': 'int32_t', 'help': 'An ``int32`` representing number of documents requested to be returned on each call to :symbol:`mongoc_change_stream_next`'}),
        ('resumeAfter', {'type': 'document', 'help': 'A ``Document`` representing the logical starting point of the change stream. The result of :symbol:`mongoc_change_stream_get_resume_token()` or the ``_id`` field  of any change received from a change stream can be used here. This option is mutually exclusive with ``startAfter`` and ``startAtOperationTime``.'}),
        ('startAfter', {'type': 'document', 'help': 'A ``Document`` representing the logical starting point of the change stream. Unlike ``resumeAfter``, this can resume notifications after an "invalidate" event. The result of :symbol:`mongoc_change_stream_get_resume_token()` or the ``_id`` field  of any change received from a change stream can be used here.  This option is mutually exclusive with ``resumeAfter`` and ``startAtOperationTime``.'}),
        ('startAtOperationTime', {'type': 'timestamp', 'help': 'A ``Timestamp``. The change stream only provides changes that occurred at or after the specified timestamp. Any command run against the server will return an operation time that can be used here. This option is mutually exclusive with ``resumeAfter`` and ``startAfter``.'}),
        ('maxAwaitTimeMS', {'type': 'int64_t', 'convert': '_mongoc_convert_int64_positive', 'help': 'An ``int64`` representing the maximum amount of time a call to :symbol:`mongoc_change_stream_next` will block waiting for data'}),
        ('fullDocument', {
            'type': 'utf8',
            'help': 'An optional UTF-8 string. Set this option to "default", '
                    '"updateLookup", "whenAvailable", or "required", If unset, '
                    'The string "default" is assumed. Set this option to '
                    '"updateLookup" to direct the change stream cursor to '
                    'lookup the most current majority-committed version of the '
                    'document associated to an update change stream event.'
        }),
        ('fullDocumentBeforeChange', {
            'type': 'utf8',
            'help': 'An optional UTF-8 string. Set this option to '
                    '"whenAvailable", "required", or "off". When unset, the '
                    'default value is "off". Similar to "fullDocument", but '
                    'returns the value of the document before the associated '
                    'change.',
        }),
        ('showExpandedEvents', { 'type': 'bool', 'help': 'Set to ``true`` to return an expanded list of change stream events. Available only on MongoDB versions >=6.0'}),
        comment_option_string_pre_4_4,
    ], fullDocument=None, fullDocumentBeforeChange=None)),

    ('mongoc_create_index_opts_t', Struct([
        write_concern_option,
        session_option,
    ], opts_name='command_opts')),

    ('mongoc_read_write_opts_t', Struct([
        read_concern_document_option,
        write_concern_option,
        session_option,
        collation_option,
        server_option,
    ])),

    # Only for documentation - we use mongoc_read_write_opts_t for real parsing.
    ('mongoc_read_opts_t', Struct([
        read_concern_document_option,
        session_option,
        collation_option,
        server_option,
    ], generate_code=False)),

    ('mongoc_write_opts_t', Struct([
        write_concern_option,
        session_option,
        collation_option,
        server_option,
    ], generate_code=False)),

    ('mongoc_gridfs_bucket_opts_t', Struct([
        ('bucketName', {'type': 'utf8', 'help': 'A UTF-8 string used as the prefix to the GridFS "chunks" and "files" collections. Defaults to "fs". The bucket name, together with the database and suffix collections must not exceed 120 characters. See the manual for `the max namespace length <https://www.mongodb.com/docs/manual/reference/limits/#Namespace-Length>`_.'}),
        ('chunkSizeBytes', {'type': 'int32_t', 'convert': '_mongoc_convert_int32_positive', 'help': 'An ``int32`` representing the chunk size. Defaults to 255KB.'}),
        write_concern_option,
        read_concern_option
    ], bucketName="fs", chunkSizeBytes=(255 * 1024))),

    ('mongoc_gridfs_bucket_upload_opts_t', Struct([
        ('chunkSizeBytes', {'type': 'int32_t', 'convert': '_mongoc_convert_int32_positive', 'help': 'An ``int32`` chunk size to use for this file. Overrides the ``chunkSizeBytes`` set on ``bucket``.'}),
        ('metadata', {'type': 'document', 'help': 'A :symbol:`bson_t` representing metadata to include with the file.'})
    ])),

    ('mongoc_aggregate_opts_t', Struct([
        read_concern_option,
        write_concern_option,
        session_option,
        bypass_option,
        collation_option,
        server_option,
        ('batchSize', {'type': 'int32_t', 'help': 'An ``int32`` representing number of documents requested to be returned on each call to :symbol:`mongoc_cursor_next`', 'check_set': True}),
        let_option,
        comment_option_string_pre_4_4,
        hint_option,
    ])),

    ('mongoc_find_and_modify_appended_opts_t', Struct([
        write_concern_option,
        session_option,
        hint_option,
        let_option,
        comment_option_since_4_4,
    ], opts_name='extra')),

    ('mongoc_count_document_opts_t', Struct([
        read_concern_document_option,
        session_option,
        collation_option,
        server_option,
        # The CRUD spec specifies `skip` and `limit` as int64_t. The server appears to accept int32 and double. The C driver accepts any bson_value_t and relies on the server to return an error for an invalid type.
        ('skip', {'type': 'bson_value_t', 'help': 'An int specifying how many documents matching the ``query`` should be skipped before counting.'}),
        ('limit', {'type': 'bson_value_t', 'help': 'An int specifying the maximum number of documents to count.'}),
        comment_option_string_pre_4_4,
        hint_option
    ])),

])

header_comment = """/**************************************************
 *
 * Generated by build/%s.
 *
 * DO NOT EDIT THIS FILE.
 *
 *************************************************/
/* clang-format off */""" % basename(__file__)


def paths(struct):
    """Sequence of path, option name, option info."""
    for option_name, info in struct.items():
        the_type = info['type']
        the_field = info.get('field', option_name)
        if the_type in opts_structs:
            # E.g., the type is mongoc_crud_opts_t. Recurse.
            sub_struct = opts_structs[the_type]
            for path, sub_option_name, sub_info in paths(sub_struct):
                yield ('%s.%s' % (the_field, path),
                       sub_option_name,
                       sub_info)
        else:
            yield the_field, option_name, info


def path_to(the_type, the_field):
    """Like "mongoc_update_one_opts->update.crud.write_concern_owned"."""
    for path, name, info in paths(opts_structs[the_type]):
        if name == the_field:
            return path

    raise ValueError(
        "No field '%s' in '%s'" % (the_field, the_type))


env = Environment(loader=FileSystemLoader(template_dir),
                  trim_blocks=True,
                  extensions=['jinja2.ext.loopcontrols'])

files = ["mongoc-opts-private.h", "mongoc-opts.c"]

for file_name in files:
    print(file_name)
    with open(joinpath(src_dir, file_name), 'w+') as f:
        t = env.get_template(file_name + ".template")
        f.write(t.render(globals()))
        f.write('\n')


def document_opts(struct, f):
    for option_name, info in struct.items():
        if info.get('internal') or info.get('hidden'):
            continue

        the_type = info['type']
        if the_type in opts_structs:
            # E.g., the type is mongoc_crud_opts_t. Recurse.
            document_opts(opts_structs[the_type], f)
            continue

        assert 'help' in info, "No 'help' for '%s'" % option_name
        f.write("* ``{option_name}``: {info[help]}\n".format(**locals()))


disclaimer = """
..
   Generated with build/generate-opts.py
   DO NOT EDIT THIS FILE

"""
for struct_name, struct in opts_structs.items():
    if not struct.generate_rst:
        continue

    name = re.sub(r'mongoc_(\w+)_t', r'\1', struct_name).replace('_', '-')
    file_name = name + '.txt'
    print(file_name)
    f = open(joinpath(doc_includes, file_name), 'w')
    f.write (disclaimer)
    f.write(
        "``%s`` may be NULL or a BSON document with additional"
        " command options:\n\n" % struct.opts_name)
    document_opts(struct, f)

    f.close()



from shrub.v3.evg_build_variant import BuildVariant
from shrub.v3.evg_command import EvgCommandType
from shrub.v3.evg_task import EvgTask, EvgTaskRef

from config_generator.components.funcs.find_cmake_latest import FindCMakeLatest

from config_generator.etc.distros import find_large_distro
from config_generator.etc.distros import make_distro_str
from config_generator.etc.distros import to_cc
from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


TAG = 'std-matrix'


# pylint: disable=line-too-long
# fmt: off
MATRIX = [
    ('debian92',          'clang',     None,   [11,   ]),
    ('debian10',          'clang',     None,   [11,   ]),
    ('debian10',          'gcc',       None,   [11, 17]),
    ('debian11',          'clang',     None,   [11,   ]),
    ('debian11',          'gcc',       None,   [11, 17]),
    ('ubuntu2004',        'clang',     None,   [11,   ]),
    ('ubuntu2004',        'gcc',       None,   [11,   ]),
    ('windows-vsCurrent', 'vs2017x64', None,   [11, 17]),
]
# fmt: on
# pylint: enable=line-too-long


class StdCompile(Function):
    name = 'std-compile'
    commands = [
        bash_exec(
            command_type=EvgCommandType.TEST,
            add_expansions_to_env=True,
            working_dir='mongoc',
            script='.evergreen/scripts/compile-std.sh',
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return StdCompile.defn()


def tasks():
    res = []

    for distro_name, compiler, arch, stds in MATRIX:
        tags = [TAG, distro_name, compiler, 'compile']

        distro = find_large_distro(distro_name)

        compile_vars = None
        compile_vars = {'CC': to_cc(compiler)}

        if arch:
            tags.append(arch)
            compile_vars.update({'MARCH': arch})

        distro_str = make_distro_str(distro_name, compiler, arch)

        for std in stds:
            with_std = {'C_STD_VERSION': std}

            task_name = f'std-c{std}-{distro_str}-compile'

            res.append(
                EvgTask(
                    name=task_name,
                    run_on=distro.name,
                    tags=tags + [f'std-c{std}'],
                    commands=[
                        FindCMakeLatest.call(),
                        StdCompile.call(vars=compile_vars | with_std)
                    ],
                )
            )

    return res


def variants():
    return [
        BuildVariant(
            name=TAG,
            display_name=TAG,
            tasks=[EvgTaskRef(name=f'.{TAG}')],
        ),
    ]



from shrub.v3.evg_command import EvgCommandType
from shrub.v3.evg_task import EvgTask

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class CheckMongocPublicHeaders(Function):
    name = 'check-headers'
    commands = [
        bash_exec(
            command_type=EvgCommandType.TEST,
            working_dir='mongoc',
            script='.evergreen/scripts/check-public-decls.sh',
        ),
        bash_exec(
            command_type=EvgCommandType.TEST,
            working_dir='mongoc',
            script='.evergreen/scripts/check-preludes.py .',
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return CheckMongocPublicHeaders.defn()


def tasks():
    return [
        EvgTask(
            name=CheckMongocPublicHeaders.name,
            commands=[CheckMongocPublicHeaders.call()],
        )
    ]



from typing import ClassVar
from typing import Mapping
from collections import ChainMap

from shrub.v3.evg_command import EvgCommand
from shrub.v3.evg_command import FunctionCall


class Function:
    name: ClassVar[str]
    commands: ClassVar[list[EvgCommand]]

    @classmethod
    def defn(cls) -> Mapping[str, list[EvgCommand]]:
        return {cls.name: cls.commands}

    @classmethod
    def default_call(cls, **kwargs) -> FunctionCall:
        return FunctionCall(func=cls.name, **kwargs)


def merge_defns(*args):
    return ChainMap(*args)



from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec

from shrub.v3.evg_command import EvgCommandType, expansions_update


class SetCacheDir(Function):
    name = 'set-cache-dir'
    commands = [
        bash_exec(
            command_type=EvgCommandType.SETUP,
            script='''\
                if [[ -n "$XDG_CACHE_DIR" ]]; then
                    cache_dir="$XDG_CACHE_DIR" # XDG Base Directory specification.
                elif [[ -n "$LOCALAPPDATA" ]]; then
                    cache_dir="$LOCALAPPDATA" # Windows.
                elif [[ -n "$USERPROFILE" ]]; then
                    cache_dir="$USERPROFILE/.cache" # Windows (fallback).
                elif [[ -d "$HOME/Library/Caches" ]]; then
                    cache_dir="$HOME/Library/Caches" # MacOS.
                elif [[ -n "$HOME" ]]; then
                    cache_dir="$HOME/.cache" # Linux-like.
                elif [[ -d ~/.cache ]]; then
                    cache_dir="~/.cache" # Linux-like (fallback).
                else
                    cache_dir="$(pwd)/.cache" # EVG task directory (fallback).
                fi

                mkdir -p "$cache_dir/mongo-c-driver" || exit
                cache_dir="$(cd "$cache_dir/mongo-c-driver" && pwd)" || exit

                printf "MONGO_C_DRIVER_CACHE_DIR: %s\\n" "$cache_dir" >|expansions.set-cache-dir.yml
            ''',
        ),
        expansions_update(
            command_type=EvgCommandType.SETUP,
            file='expansions.set-cache-dir.yml'
        ),
    ]


def functions():
    return SetCacheDir.defn()



# Copyright 2009-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
from collections import OrderedDict as OD
from typing import Any, Iterable, Mapping, MutableMapping, MutableSequence, Sequence, Union

Scalar = Union[str, bool, int, None, float]
"YAML simple schema scalar types"
ValueSequence = Sequence["Value"]
"Sequence of YAML simple values"
MutableValueArray = MutableSequence["Value"]
"A mutable sequence of JSON values"
ValueMapping = Mapping[Scalar, "Value"]
"A YAML mapping type (arbitrary scalars as keys)"
MutableValueMapping = MutableMapping[Scalar, "Value"]
"A mutable YAML mapping type"
Value = Union[ValueSequence, ValueMapping, Scalar]
"Any YAML simple value"
MutableValue = Union[MutableValueMapping, MutableValueArray, Scalar]
"Any YAML simple value, which may be a mutable sequence or map"

ValueOrderedDict = OD[Scalar, Value]
"An OrderedDict of YAML values"


try:
    import yaml
    import yamlordereddictloader  # type: ignore
except ImportError:
    sys.stderr.write("try 'poetry install --with=dev'\n")
    raise


class ConfigObject(object):
    @property
    def name(self) -> str:
        return "UNSET"

    def to_dict(self) -> Value:
        return OD([("name", self.name)])


# We want legible YAML tasks:
#
#     - name: debug-compile
#       tags: [zlib, snappy, compression, openssl]
#       commands:
#       - command: shell.exec
#         params:
#           script: |-
#             set -o errexit
#             ...
#
# Write values compactly except multiline strings, which use "|" style. Write
# tag sets as lists.


class _Dumper(yamlordereddictloader.Dumper):
    def __init__(self, *args: Value, **kwargs: Value):
        super().__init__(*args, **kwargs)  # type: ignore
        self.add_representer(set, type(self).represent_set)
        # Use "multi_representer" to represent all subclasses of ConfigObject.
        self.add_multi_representer(ConfigObject, type(self).represent_config_object)

    def represent_scalar(self, tag: str, value: Value, style: str | None = None) -> yaml.ScalarNode:
        if isinstance(value, (str)) and "\n" in value:
            style = "|"
        return super().represent_scalar(tag, value, style)  # type: ignore

    def represent_set(self, data: Iterable[Value]) -> yaml.MappingNode:
        return super().represent_list(sorted(set(data)))  # type: ignore

    def represent_config_object(self, obj: ConfigObject) -> yaml.Node:
        d = obj.to_dict()
        return super().represent_data(d)  # type: ignore


def yaml_dump(obj: Any):
    return yaml.dump(obj, Dumper=_Dumper, default_flow_style=False)


def generate(config: Any, path: str):
    """Dump config to a file as YAML.
    config is a dict, preferably an OrderedDict. path is a file path.
    """
    f = open(path, "w+")
    f.write(
        """####################################
# Evergreen configuration
#
# Generated with evergreen_config_generator from
# github.com/mongodb-labs/drivers-evergreen-tools
#
# DO NOT EDIT THIS FILE
#
####################################
"""
    )
    f.write(yaml_dump(config))



from __future__ import annotations

import enum
import struct
from typing import Final, Iterable, NamedTuple, Sequence


def generate():
    simple_oid = OID((1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12))
    ITEMS: list[LineItem] = [
        Comment("This file is GENERATED! DO NOT MODIFY!"),
        Comment("Instead, modify the content of make-dicts.py"),
        Line(),
        Comment("Random values"),
        Entry("int32_1729", encode_value(1729)),
        Entry("int64_1729", struct.pack("<q", 1729)),
        Entry("prefixed_string", make_string("prefixed-string")),
        Entry("empty_obj", wrap_obj(b"")),
        Entry("c_string", make_cstring("null-terminated-string")),
        Line(),
        Comment("Elements"),
        Entry("null_elem", element("N", None)),
        Entry("undef_elem", element("U", Undefined)),
        Entry("string_elem", element("S", "string-value")),
        Entry("empty_bin_elem", element("Bg", Binary(0, b""))),
        Entry("empty_regex_elem", element("Rx0", Regex("", ""))),
        Entry("simple_regex_elem", element("Rx1", Regex("foo", "ig"))),
        Entry("encrypted_bin_elem", element("Be", Binary(6, b"meow"))),
        Entry("empty_obj_elem", element("Obj0", Doc())),
        Entry(
            "code_w_s_elem",
            element("Clz", CodeWithScope("void 0;", Doc([Elem("foo", "bar")]))),
        ),
        Entry("code_elem", element("Js", Code("() => 0;"))),
        Entry("symbol_elem", element("Sym", Symbol("symbol"))),
        Entry("oid_elem", element("OID", simple_oid)),
        Entry("dbpointer_elem", element("dbp", DBPointer(String("db"), simple_oid))),
        Line(),
        Comment("Embedded nul"),
        Comment("This string contains an embedded null, which is abnormal but valid"),
        Entry("string_with_null", element("S0", "string\0value")),
        Comment("This regex has an embedded null, which is invalid"),
        Entry("bad_regex_elem", element("RxB", Regex("f\0oo", "ig"))),
        Comment("This element's key contains an embedded null, which is invalid"),
        Entry("bad_key_elem", element("foo\0bar", "string")),
        Line(),
        Comment("Objects"),
        Entry("obj_with_string", wrap_obj(element("single-elem", "foo"))),
        Entry("obj_with_null", wrap_obj(element("null", None))),
        Entry("obj_missing_term", wrap_obj(b"")[:-1]),
    ]

    for it in ITEMS:
        emit(it)


BytesIter = bytes | Iterable["BytesIter"]


def flatten(b: BytesIter) -> bytes:
    if isinstance(b, bytes):
        return b
    else:
        return b"".join(map(flatten, b))


def len_prefix(b: BytesIter) -> bytes:
    """Prepend an i32le byte-length prefix to a set of bytes"""
    b = flatten(b)
    length = len(b)
    return encode_value(length) + b


def make_cstring(s: str) -> bytes:
    """Encode a UTF-8 string and append a null terminator"""
    return s.encode("utf-8") + b"\0"


def make_string(s: str) -> bytes:
    """Create a length-prefixed string byte sequence"""
    return len_prefix(make_cstring(s))


def wrap_obj(items: BytesIter) -> bytes:
    """Wrap a sequence of bytes as if a BSON object (adds a header and trailing nul)"""
    bs = flatten(items)
    header = len(bs) + 5
    return encode_value(header) + bs + b"\0"


class UndefinedType:
    def __bytes__(self) -> bytes:
        return b""


class Binary(NamedTuple):
    tag: int
    data: bytes

    def __bytes__(self) -> bytes:
        return encode_value(len(self.data)) + bytes([self.tag]) + self.data


class OID(NamedTuple):
    octets: tuple[int, int, int, int, int, int, int, int, int, int, int, int]

    def __bytes__(self) -> bytes:
        return bytes(self.octets)


class DBPointer(NamedTuple):
    db: String
    oid: OID

    def __bytes__(self) -> bytes:
        return self.db.__bytes__() + self.oid.__bytes__()


class Regex(NamedTuple):
    rx: str
    opts: str

    def __bytes__(self) -> bytes:
        return make_cstring(self.rx) + make_cstring(self.opts)


class Elem(NamedTuple):
    key: str
    val: ValueType

    def __bytes__(self) -> bytes:
        return element(self.key, self.val)


class Doc(NamedTuple):
    items: Sequence[Elem] = ()

    def __bytes__(self) -> bytes:
        return wrap_obj((e.__bytes__() for e in self.items))


class String(NamedTuple):
    s: str

    def __bytes__(self) -> bytes:
        return make_string(self.s)


class Symbol(String):
    pass


class Code(String):
    pass


class CodeWithScope(NamedTuple):
    code: str
    scope: Doc

    def __bytes__(self) -> bytes:
        string = make_string(self.code)
        doc = self.scope.__bytes__()
        length = len(string) + len(doc) + 4
        return encode_value(length) + string + doc


Undefined: Final = UndefinedType()
ValueType = (
    int
    | str
    | float
    | None
    | UndefinedType
    | Binary
    | Regex
    | Doc
    | CodeWithScope
    | String
    | Symbol
    | Code
    | OID
    | DBPointer
)


def encode_value(val: ValueType) -> bytes:
    match val:
        case int(n):
            return struct.pack("<i", n)
        case str(s):
            return make_string(s)
        case float(f):
            return struct.pack("<d", f)
        case (
            Doc()
            | Binary()
            | Regex()
            | UndefinedType()
            | CodeWithScope()
            | Code()
            | String()
            | Symbol()
            | DBPointer()
            | OID()
        ) as d:
            return d.__bytes__()
        case None:
            return b""


class Tag(enum.Enum):
    EOD = 0
    Double = 1
    UTF8 = 2
    Doc = 3
    Binary = 5
    Undefined = 6
    OID = 7
    Null = 10
    Regex = 11
    DBPointer = 12
    Code = 13
    Symbol = 14
    CodeWithScope = 15
    Int32 = 16
    Int64 = 18


def element(key: str, value: ValueType, *, type: None | Tag = None) -> bytes:
    if type is not None:
        return flatten([bytes([type.value]), make_cstring(key), encode_value(value)])

    match value:
        case int():
            type = Tag.Int32

        case float():
            type = Tag.Double
        case None:
            type = Tag.Null
        case UndefinedType():
            type = Tag.Undefined
        case Binary():
            type = Tag.Binary
        case Regex():
            type = Tag.Regex
        case Doc():
            type = Tag.Doc
        case CodeWithScope():
            type = Tag.CodeWithScope
        case Code():
            type = Tag.Code
        case Symbol():
            type = Tag.Symbol
        case str() | String():  # Must appear after Code()/Symbol()
            type = Tag.UTF8
        case OID():
            type = Tag.OID
        case DBPointer():
            type = Tag.DBPointer
    return element(key, value, type=type)


class Entry(NamedTuple):
    key: str
    "The key for the entry. Only for human readability"
    value: bytes
    "The arbitrary bytes that make up the entry"


class Comment(NamedTuple):
    txt: str


class Line(NamedTuple):
    txt: str = ""


LineItem = Entry | Comment | Line


def escape(b: bytes) -> Iterable[str]:
    s = b.decode("ascii", "backslashreplace")
    for u8 in b:
        s = chr(u8)  # 0 <= u8 and u8 <= 255
        if s.isascii() and s.isprintable():
            yield s
            continue
        # Byte is not valid ASCII, or is not a printable char
        yield f"\\x{u8:0>2x}"


def emit(item: LineItem):
    match item:
        case Line(t):
            print(t)
        case Comment(txt):
            print(f"# {txt}")
        case Entry(key, val):
            s = "".join(escape(val))
            s = s.replace('"', r"\x22")
            print(f'{key}="{s}"')


if __name__ == "__main__":
    generate()



from shrub.v3.evg_command import EvgCommandType
from shrub.v3.evg_command import expansions_update

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class FetchDET(Function):
    name = 'fetch-det'
    commands = [
        bash_exec(
            command_type=EvgCommandType.SETUP,
            script='''\
                if [[ ! -d drivers-evergreen-tools ]]; then
                    git clone --depth=1 https://github.com/mongodb-labs/drivers-evergreen-tools.git
                fi
            ''',
        ),

        # Make shell scripts executable.
        bash_exec(
            command_type=EvgCommandType.SETUP,
            working_dir="drivers-evergreen-tools",
            script='find .evergreen -type f -name "*.sh" -exec chmod +rx "{}" \;',
        ),

        # python is used frequently enough by many tasks that it is worth
        # running find_python3 once here and reusing the result.
        bash_exec(
            command_type=EvgCommandType.SETUP,
            script='''\
                set -o errexit
                . drivers-evergreen-tools/.evergreen/find-python3.sh
                echo "PYTHON3_BINARY: $(find_python3)" >|python3_binary.yml
            ''',
        ),
        expansions_update(
            command_type=EvgCommandType.SETUP,
            file='python3_binary.yml',
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return FetchDET.defn()



import subprocess
import sys


def generate():
    subprocess.run(
        args=[
            sys.executable,
            '.evergreen/legacy_config_generator/generate-evergreen-config.py',
        ],
        check=True,
    )



from shrub.v3.evg_build_variant import BuildVariant
from shrub.v3.evg_command import EvgCommandType
from shrub.v3.evg_task import EvgTask
from shrub.v3.evg_task import EvgTaskRef

from config_generator.etc.distros import find_large_distro
from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


TAG = "clang-format"

DISTROS = [
    find_large_distro("ubuntu2204").name,
    find_large_distro("ubuntu2004").name,
]


class ClangFormat(Function):
    name = TAG
    commands = [
        bash_exec(
            command_type=EvgCommandType.SETUP,
            working_dir="mongoc",
            script="./tools/poetry.sh install --with=dev"
        ),
        bash_exec(
            command_type=EvgCommandType.TEST,
            working_dir="mongoc",
            env={
                "DRYRUN": "1",
            },
            script="./tools/poetry.sh run .evergreen/scripts/clang-format-all.sh",
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return ClangFormat.defn()


def tasks():
    yield EvgTask(
        name=TAG,
        tags=[TAG],
        commands=[ClangFormat.call()],
    )


def variants():
    return [
        BuildVariant(
            name=TAG,
            display_name=TAG,
            run_on=DISTROS,
            tasks=[EvgTaskRef(name=f'.{TAG}')],
        ),
    ]



from shrub.v3.evg_build_variant import BuildVariant
from shrub.v3.evg_command import EvgCommandType
from shrub.v3.evg_task import EvgTask, EvgTaskRef

from config_generator.components.funcs.fetch_det import FetchDET
from config_generator.components.funcs.find_cmake_latest import FindCMakeLatest
from config_generator.components.funcs.run_simple_http_server import RunSimpleHTTPServer
from config_generator.etc.utils import bash_exec

def tasks():
    return [
        EvgTask(
            name="mock-server-test",
            run_on="ubuntu2204-small",
            commands=[
                # Call fetch-det to define PYTHON3_BINARY expansion required for run-simple-http-server.
                FetchDET.call(),
                RunSimpleHTTPServer.call(),
                FindCMakeLatest.call(),
                bash_exec(
                    command_type=EvgCommandType.TEST,
                    add_expansions_to_env=True,
                    working_dir='mongoc',
                    script='.evergreen/scripts/compile.sh',
                ),
                bash_exec(
                    command_type=EvgCommandType.TEST,
                    working_dir='mongoc',
                    script='.evergreen/scripts/run-mock-server-tests.sh',
                )
            ],
        )
    ]

def variants():
    return [
        BuildVariant(
            name="mock-server-test",
            display_name="Mock Server Test",
            tasks=[EvgTaskRef(name='mock-server-test')],
            expansions={
                'CC': 'gcc',
                'ASAN': 'on',
                'CFLAGS': '-fno-omit-frame-pointer',
                'EXTRA_CONFIGURE_FLAGS': '-DENABLE_EXTRA_ALIGNMENT=OFF',
                'SANITIZE': 'address,undefined',
            }
        ),
    ]



from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class StopMongoOrchestration(Function):
    name = 'stop-mongo-orchestration'
    commands = [
        bash_exec(
            script='''\
                if [[ -d MO ]]; then
                    cd MO && mongo-orchestration stop
                fi
            '''
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return StopMongoOrchestration.defn()



from shrub.v3.evg_build_variant import BuildVariant
from shrub.v3.evg_task import EvgTaskRef

from config_generator.etc.function import merge_defns
from config_generator.etc.compile import generate_compile_tasks

from config_generator.etc.sasl.compile import CompileCommon
from config_generator.etc.sasl.test import generate_test_tasks


SSL = 'nossl'
TAG = f'sasl-matrix-{SSL}'


# pylint: disable=line-too-long
# fmt: off
COMPILE_MATRIX = [
    ('ubuntu2004',        'gcc',         None,   ['off']),
    ('windows-vsCurrent', 'vs2017x64',   None,   ['off']),
    ('debian10',          'gcc',         None,   ['off']),
]

TEST_MATRIX = [
    ('ubuntu2004', 'gcc', None, 'off', ['noauth'], ['server', 'replica', 'sharded'], ['4.4', '5.0', '6.0', '7.0', '8.0', 'latest']),

    # Test 4.2 with Debian 10 since 4.2 does not ship on Ubuntu 20.04+.
    ('debian10',   'gcc', None, 'off', ['noauth'], ['server', 'replica', 'sharded'], ['4.2']), 
]
# fmt: on
# pylint: enable=line-too-long


class NoSSLCompileCommon(CompileCommon):
    ssl = 'OFF'


class SaslOffNoSSLCompile(NoSSLCompileCommon):
    name = 'sasl-off-nossl-compile'
    commands = NoSSLCompileCommon.compile_commands(sasl='OFF')


def functions():
    return merge_defns(
        SaslOffNoSSLCompile.defn(),
    )


def tasks():
    res = []

    SASL_TO_FUNC = {
        'off': SaslOffNoSSLCompile,
    }

    res += generate_compile_tasks(SSL, TAG, SASL_TO_FUNC, COMPILE_MATRIX)
    res += generate_test_tasks(SSL, TAG, TEST_MATRIX)

    return res


def variants():
    expansions = {}

    return [
        BuildVariant(
            name=TAG,
            display_name=TAG,
            tasks=[EvgTaskRef(name=f'.{TAG}')],
            expansions=expansions,
        ),
    ]



#!/usr/bin/env python3
#
# Copyright 2009-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Check that public libbson/libmongoc headers all include the prelude line.
"""
import sys
from pathlib import Path

if len(sys.argv) != 2:
    print("Usage: python check-preludes.py <mongo-c-driver directory>")
    sys.exit(1)

MONGOC_PREFIX = Path("src/libmongoc/src/mongoc")
BSON_PREFIX = Path("src/libbson/src/bson")
COMMON_PREFIX = Path("src/common/src")

checks = [
    {
        "name": "libmongoc",
        "headers": list(MONGOC_PREFIX.glob("mongoc-*.h")),
        "exclusions": [
            MONGOC_PREFIX / "mongoc-prelude.h",
            MONGOC_PREFIX / "mongoc.h",
        ],
        "include": '#include <mongoc/mongoc-prelude.h>',
    },
    {
        "name": "libbson",
        "headers": list(BSON_PREFIX.glob("*.h")),
        "exclusions": [
            BSON_PREFIX / "bson-prelude.h",
            BSON_PREFIX / "bson.h",
        ],
        "include": "#include <bson/bson-prelude.h>",
    },
    {
        "name": "common",
        "headers": list(COMMON_PREFIX.glob("*.h")),
        "exclusions": [COMMON_PREFIX / "common-prelude.h"],
        "include": '#include <common-prelude.h>',
    },
]

for check in checks:
    NAME = check["name"]
    print(f"Checking headers for {NAME}")
    assert len(check["headers"]) > 0
    for header in check["headers"]:
        if header in check["exclusions"]:
            continue
        lines = Path(header).read_text(encoding="utf-8").splitlines()
        if check["include"] not in lines:
            print(f"{header} did not include prelude")
            sys.exit(1)

print("All checks passed")



from shrub.v3.evg_build_variant import BuildVariant
from shrub.v3.evg_command import EvgCommandType
from shrub.v3.evg_task import EvgTask, EvgTaskRef

from config_generator.components.funcs.find_cmake_latest import FindCMakeLatest

from config_generator.etc.distros import find_large_distro
from config_generator.etc.distros import make_distro_str
from config_generator.etc.distros import to_cc
from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec

SSL = 'openssl-static'
TAG = f'{SSL}-matrix'


# pylint: disable=line-too-long
# fmt: off
MATRIX = [
  ('debian92',   'gcc', None),
  ('debian10',   'gcc', None),
  ('debian11',   'gcc', None),
  ('ubuntu2004', 'gcc', None),
]
# fmt: on
# pylint: enable=line-too-long


class StaticOpenSSLCompile(Function):
    name = 'openssl-static-compile'
    commands = [
        bash_exec(
            command_type=EvgCommandType.TEST,
            add_expansions_to_env=True,
            working_dir='mongoc',
            script='.evergreen/scripts/compile-openssl-static.sh',
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return StaticOpenSSLCompile.defn()


def tasks():
    res = []

    for distro_name, compiler, arch, in MATRIX:
        tags = [TAG, distro_name, compiler]

        distro = find_large_distro(distro_name)

        compile_vars = None
        compile_vars = {'CC': to_cc(compiler)}

        if arch:
            tags.append(arch)
            compile_vars.update({'MARCH': arch})

        distro_str = make_distro_str(distro_name, compiler, arch)

        task_name = f'openssl-static-compile-{distro_str}'

        res.append(
            EvgTask(
                name=task_name,
                run_on=distro.name,
                tags=tags,
                commands=[
                    FindCMakeLatest.call(),
                    StaticOpenSSLCompile.call(vars=compile_vars),
                ],
            )
        )

    return res


def variants():
    return [
        BuildVariant(
            name=TAG,
            display_name=TAG,
            tasks=[EvgTaskRef(name=f'.{TAG}')],
        ),
    ]



import functools
import json
import sys
import time
import traceback
from pathlib import Path

import bottle
from bottle import Bottle, HTTPResponse

kms_provider = Bottle(autojson=True)
"""A mock server for Azure IMDS and GCP metadata"""

from typing import TYPE_CHECKING, Any, Callable, Iterable, cast, overload

if not TYPE_CHECKING:
    from bottle import request
else:
    from typing import Protocol

    class _RequestParams(Protocol):

        def __getitem__(self, key: str) -> str:
            ...

        @overload
        def get(self, key: str) -> 'str | None':
            ...

        @overload
        def get(self, key: str, default: str) -> str:
            ...

    class _HeadersDict(dict[str, str]):

        def raw(self, key: str) -> 'bytes | None':
            ...

    class _Request(Protocol):

        @property
        def query(self) -> _RequestParams:
            ...

        @property
        def params(self) -> _RequestParams:
            ...

        @property
        def headers(self) -> _HeadersDict:
            ...

    request = cast('_Request', None)


def parse_qs(qs: str) -> 'dict[str, str]':
    # Re-use the bottle.py query string parser. It's a private function, but
    # we're using a fixed version of Bottle.
    return dict(bottle._parse_qsl(qs))  # type: ignore


_HandlerFuncT = Callable[
    [],
    'None|str|bytes|dict[str, Any]|bottle.BaseResponse|Iterable[bytes|str]']


def handle_asserts(fn: _HandlerFuncT) -> _HandlerFuncT:
    "Convert assertion failures into HTTP 400s"

    @functools.wraps(fn)
    def wrapped():
        try:
            return fn()
        except AssertionError as e:
            traceback.print_exc()
            return bottle.HTTPResponse(status=400,
                                       body=json.dumps({'error':
                                                        list(e.args)}))

    return wrapped


def test_params() -> 'dict[str, str]':
    return parse_qs(request.headers.get('X-MongoDB-HTTP-TestParams', ''))

@kms_provider.get('/computeMetadata/v1/instance/service-accounts/default/token')
@handle_asserts
def get_gcp_token():
    metadata_header = request.headers.get("Metadata-Flavor")
    assert metadata_header == 'Google'

    case = test_params().get('case')
    print('Case is:', case)
    if case == '404': 
        return HTTPResponse(status=404)
    
    if case == 'bad-json':
        return b'{"access-token": }'
    
    if case == 'empty-json':
        return b'{}'

    if case == 'giant':
        return _gen_giant()

    if case == 'slow':
        return _slow()

    assert case in (None, ''), 'Unknown HTTP test case "{}"'.format(case)
    
    return {
        'access_token' : 'google-cookie',
        'token_type' : 'Bearer'
    }

@kms_provider.get('/metadata/identity/oauth2/token')
@handle_asserts
def get_oauth2_token():
    api_version = request.query['api-version']
    assert api_version == '2018-02-01', 'Only api-version=2018-02-01 is supported'
    resource = request.query['resource']
    assert resource == 'https://vault.azure.net', 'Only https://vault.azure.net is supported'

    case = test_params().get('case')
    print('Case is:', case)
    if case == '404':
        return HTTPResponse(status=404)

    if case == '500':
        return HTTPResponse(status=500)

    if case == 'bad-json':
        return b'{"key": }'

    if case == 'empty-json':
        return b'{}'

    if case == 'giant':
        return _gen_giant()

    if case == 'slow':
        return _slow()

    assert case in (None, ''), 'Unknown HTTP test case "{}"'.format(case)

    return {
        'access_token': 'magic-cookie',
        'expires_in': '70',
        'token_type': 'Bearer',
        'resource': 'https://vault.azure.net',
    }


def _gen_giant() -> Iterable[bytes]:
    "Generate a giant message"
    yield b'{ "item": ['
    for _ in range(1024 * 256):
        yield (b'null, null, null, null, null, null, null, null, null, null, '
               b'null, null, null, null, null, null, null, null, null, null, '
               b'null, null, null, null, null, null, null, null, null, null, '
               b'null, null, null, null, null, null, null, null, null, null, ')
    yield b' null ] }'
    yield b'\n'


def _slow() -> Iterable[bytes]:
    "Generate a very slow message"
    yield b'{ "item": ['
    for _ in range(1000):
        yield b'null, '
        time.sleep(1)
    yield b' null ] }'


if __name__ == '__main__':
    print(
        'RECOMMENDED: Run this script using bottle.py (e.g. [{} {}/bottle.py fake_kms_provider_server:kms_provider])'
        .format(sys.executable,
                Path(__file__).resolve().parent))
    kms_provider.run()



# -*- coding: utf-8 -*-
from docutils import nodes
import os
import sys

# Import common docs config.
this_path = os.path.dirname(__file__)
sys.path.append(os.path.normpath(os.path.join(this_path, '../')))

from mongoc_common import *

with open(this_path + '/../../../VERSION_CURRENT') as vc:
    release = version = vc.read()

# -- General configuration ------------------------------------------------
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'

# General information about the project.
project = u'mongoc.org'
copyright = u'2009-present, MongoDB, Inc.'
author = u'MongoDB, Inc'
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# Support :download-link:`bson` or :download-link:`mongoc`.
def download_link(typ, rawtext, text, lineno, inliner, options={}, content=[]):
    if text == "mongoc":
        lib = "mongo-c-driver"
    else:
        raise ValueError(
            "download link must be mongoc, not \"%s\"" % text)

    title = "%s-%s" % (lib, version)
    url = ("https://github.com/mongodb/mongo-c-driver/releases/tag/%(version)s") % {
              "version": version
          }

    pnode = nodes.reference(title, title, internal=False, refuri=url)
    return [pnode], []

def setup(app):
    mongoc_common_setup(app)

    app.add_role('download-link', download_link)

# -- Options for HTML output ----------------------------------------------

html_theme = 'furo'
html_title = html_shorttitle = 'MongoDB C Driver %s' % version
# html_favicon = None
html_use_smartypants = False
html_show_sourcelink = False
html_use_index = False
rst_prolog = rf"""

.. _mongodb_docs_cdriver: https://www.mongodb.com/docs/languages/c/c-driver/current/

"""

html_sidebars = {
    '**': []
}

# Note: http://www.sphinx-doc.org/en/1.5.1/config.html#confval-html_copy_source
# This will degrade the Javascript quicksearch if we ever use it.
html_copy_source = False



from typing import ClassVar

from shrub.v3.evg_command import EvgCommand
from shrub.v3.evg_command import EvgCommandType
from shrub.v3.evg_command import expansions_update
from shrub.v3.evg_command import KeyValueParam

from config_generator.etc.utils import bash_exec

from config_generator.etc.function import Function


class CompileCommon(Function):
    ssl: ClassVar[str | None]

    @classmethod
    def compile_commands(cls, sasl=None) -> list[EvgCommand]:
        updates = []

        if cls.ssl:
            updates.append(KeyValueParam(key='SSL', value=cls.ssl))

        if sasl:
            updates.append(KeyValueParam(key='SASL', value=sasl))

        return [
            expansions_update(updates=updates),
            bash_exec(
                command_type=EvgCommandType.TEST,
                script='EXTRA_CONFIGURE_FLAGS="-DENABLE_PIC=ON ${EXTRA_CONFIGURE_FLAGS}" .evergreen/scripts/compile.sh',
                working_dir='mongoc',
                add_expansions_to_env=True,
                env={
                    'COMPILE_LIBMONGOCRYPT': 'ON',
                },
            ),
        ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)



#
# This file has been modified in 2019 by MongoDB Inc.
#

# OCSPBuilder is derived from https://github.com/wbond/ocspbuilder
# OCSPResponder is derived from https://github.com/threema-ch/ocspresponder

# Copyright (c) 2015-2018 Will Bond <will@wbond.net>

# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Copyright 2016 Threema GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals, division, absolute_import, print_function

import logging
import base64
import inspect
import re
import enum
import sys
import textwrap
from datetime import datetime, timezone, timedelta
from typing import Callable, Tuple, Optional

from asn1crypto import x509, keys, core, ocsp
from asn1crypto.ocsp import OCSPRequest, OCSPResponse
from oscrypto import asymmetric
from flask import Flask, request, Response

__version__ = '0.10.2'
__version_info__ = (0, 10, 2)

logger = logging.getLogger(__name__)

if sys.version_info < (3,):
    byte_cls = str
else:
    byte_cls = bytes

def _pretty_message(string, *params):
    """
    Takes a multi-line string and does the following:
     - dedents
     - converts newlines with text before and after into a single line
     - strips leading and trailing whitespace
    :param string:
        The string to format
    :param *params:
        Params to interpolate into the string
    :return:
        The formatted string
    """

    output = textwrap.dedent(string)

    # Unwrap lines, taking into account bulleted lists, ordered lists and
    # underlines consisting of = signs
    if output.find('\n') != -1:
        output = re.sub('(?<=\\S)\n(?=[^ \n\t\\d\\*\\-=])', ' ', output)

    if params:
        output = output % params

    output = output.strip()

    return output


def _type_name(value):
    """
    :param value:
        A value to get the object name of
    :return:
        A unicode string of the object name
    """

    if inspect.isclass(value):
        cls = value
    else:
        cls = value.__class__
    if cls.__module__ in set(['builtins', '__builtin__']):
        return cls.__name__
    return '%s.%s' % (cls.__module__, cls.__name__)

def _writer(func):
    """
    Decorator for a custom writer, but a default reader
    """

    name = func.__name__
    return property(fget=lambda self: getattr(self, '_%s' % name), fset=func)


class OCSPResponseBuilder(object):

    _response_status = None
    _certificate = None
    _certificate_status = None
    _revocation_date = None
    _certificate_issuer = None
    _hash_algo = None
    _key_hash_algo = None
    _nonce = None
    _this_update = None
    _next_update = None
    _response_data_extensions = None
    _single_response_extensions = None

    def __init__(self, response_status, certificate_status_list=[], revocation_date=None):
        """
        Unless changed, responses will use SHA-256 for the signature,
        and will be valid from the moment created for one week.
        :param response_status:
            A unicode string of OCSP response type:
            - "successful" - when the response includes information about the certificate
            - "malformed_request" - when the request could not be understood
            - "internal_error" - when an internal error occured with the OCSP responder
            - "try_later" - when the OCSP responder is temporarily unavailable
            - "sign_required" - when the OCSP request must be signed
            - "unauthorized" - when the responder is not the correct responder for the certificate
        :param certificate_list:
            A list of tuples with certificate serial number and certificate status objects.
            certificate_status:
                A unicode string of the status of the certificate. Only required if
                the response_status is "successful".
                - "good" - when the certificate is in good standing
                - "revoked" - when the certificate is revoked without a reason code
                - "key_compromise" - when a private key is compromised
                - "ca_compromise" - when the CA issuing the certificate is compromised
                - "affiliation_changed" - when the certificate subject name changed
                - "superseded" - when the certificate was replaced with a new one
                - "cessation_of_operation" - when the certificate is no longer needed
                - "certificate_hold" - when the certificate is temporarily invalid
                - "remove_from_crl" - only delta CRLs - when temporary hold is removed
                - "privilege_withdrawn" - one of the usages for a certificate was removed
                - "unknown" - the responder doesn't know about the certificate being requested
        :param revocation_date:
            A datetime.datetime object of when the certificate was revoked, if
            the response_status is "successful" and the certificate status is
            not "good" or "unknown".
        """
        self._response_status = response_status
        self._certificate_status_list = certificate_status_list
        self._revocation_date = revocation_date

        self._key_hash_algo = 'sha1'
        self._hash_algo = 'sha256'
        self._response_data_extensions = {}
        self._single_response_extensions = {}

    @_writer
    def nonce(self, value):
        """
        The nonce that was provided during the request.
        """

        if not isinstance(value, byte_cls):
            raise TypeError(_pretty_message(
                '''
                nonce must be a byte string, not %s
                ''',
                _type_name(value)
            ))

        self._nonce = value

    @_writer
    def certificate_issuer(self, value):
        """
        An asn1crypto.x509.Certificate object of the issuer of the certificate.
        This should only be set if the OCSP responder is not the issuer of
        the certificate, but instead a special certificate only for OCSP
        responses.
        """

        if value is not None:
            is_oscrypto = isinstance(value, asymmetric.Certificate)
            if not is_oscrypto and not isinstance(value, x509.Certificate):
                raise TypeError(_pretty_message(
                    '''
                    certificate_issuer must be an instance of
                    asn1crypto.x509.Certificate or
                    oscrypto.asymmetric.Certificate, not %s
                    ''',
                    _type_name(value)
                ))

            if is_oscrypto:
                value = value.asn1

        self._certificate_issuer = value

    @_writer
    def next_update(self, value):
        """
        A datetime.datetime object of when the response may next change. This
        should only be set if responses are cached. If responses are generated
        fresh on every request, this should not be set.
        """

        if not isinstance(value, datetime):
            raise TypeError(_pretty_message(
                '''
                next_update must be an instance of datetime.datetime, not %s
                ''',
                _type_name(value)
            ))

        self._next_update = value

    def build(self, responder_private_key=None, responder_certificate=None):
        """
        Validates the request information, constructs the ASN.1 structure and
        signs it.
        The responder_private_key and responder_certificate parameters are onlystr
        required if the response_status is "successful".
        :param responder_private_key:
            An asn1crypto.keys.PrivateKeyInfo or oscrypto.asymmetric.PrivateKey
            object for the private key to sign the response with
        :param responder_certificate:
            An asn1crypto.x509.Certificate or oscrypto.asymmetric.Certificate
            object of the certificate associated with the private key
        :return:
            An asn1crypto.ocsp.OCSPResponse object of the response
        """
        if self._response_status != 'successful':
            return ocsp.OCSPResponse({
                'response_status': self._response_status
            })

        is_oscrypto = isinstance(responder_private_key, asymmetric.PrivateKey)
        if not isinstance(responder_private_key, keys.PrivateKeyInfo) and not is_oscrypto:
            raise TypeError(_pretty_message(
                '''
                responder_private_key must be an instance ofthe c
                asn1crypto.keys.PrivateKeyInfo or
                oscrypto.asymmetric.PrivateKey, not %s
                ''',
                _type_name(responder_private_key)
            ))

        cert_is_oscrypto = isinstance(responder_certificate, asymmetric.Certificate)
        if not isinstance(responder_certificate, x509.Certificate) and not cert_is_oscrypto:
            raise TypeError(_pretty_message(
                '''
                responder_certificate must be an instance of
                asn1crypto.x509.Certificate or
                oscrypto.asymmetric.Certificate, not %s
                ''',
                _type_name(responder_certificate)
            ))

        if cert_is_oscrypto:
            responder_certificate = responder_certificate.asn1

        if self._certificate_status_list is None:
            raise ValueError(_pretty_message(
                '''
                certificate_status_list must be set if the response_status is
                "successful"
                '''
            ))

        def _make_extension(name, value):
            return {
                'extn_id': name,
                'critical': False,
                'extn_value': value
            }

        responses = []
        for serial, status in self._certificate_status_list:
            response_data_extensions = []
            single_response_extensions = []
            for name, value in self._response_data_extensions.items():
                response_data_extensions.append(_make_extension(name, value))
            if self._nonce:
                response_data_extensions.append(
                    _make_extension('nonce', self._nonce)
                )

            if not response_data_extensions:
                response_data_extensions = None

            for name, value in self._single_response_extensions.items():
                single_response_extensions.append(_make_extension(name, value))

            if self._certificate_issuer:
                single_response_extensions.append(
                    _make_extension(
                        'certificate_issuer',
                        [
                            x509.GeneralName(
                                name='directory_name',
                                value=self._certificate_issuer.subject
                            )
                        ]
                    )
                )

            if not single_response_extensions:
                single_response_extensions = None

            responder_key_hash = getattr(responder_certificate.public_key, self._key_hash_algo)

            if status == 'good':
                cert_status = ocsp.CertStatus(
                    name='good',
                    value=core.Null()
                )
            elif status == 'unknown':
                cert_status = ocsp.CertStatus(
                    name='unknown',
                    value=core.Null()
                )
            else:
                reason = status if status != 'revoked' else 'unspecified'
                cert_status = ocsp.CertStatus(
                    name='revoked',
                    value={
                        'revocation_time': self._revocation_date,
                        'revocation_reason': reason,
                    }
                )

            issuer = self._certificate_issuer if self._certificate_issuer else responder_certificate

            produced_at = datetime.now(timezone.utc).replace(microsecond=0)

            if self._this_update is None:
                self._this_update = produced_at

            if self._next_update is None:
                self._next_update = (self._this_update + timedelta(days=7)).replace(microsecond=0)

            response = {
                    'cert_id': {
                        'hash_algorithm': {
                            'algorithm': self._key_hash_algo
                        },
                        'issuer_name_hash': getattr(issuer.subject, self._key_hash_algo),
                        'issuer_key_hash': getattr(issuer.public_key, self._key_hash_algo),
                        'serial_number': serial,
                    },
                    'cert_status': cert_status,
                    'this_update': self._this_update,
                    'next_update': self._next_update,
                    'single_extensions': single_response_extensions
                }
            responses.append(response)

        response_data = ocsp.ResponseData({
            'responder_id': ocsp.ResponderId(name='by_key', value=responder_key_hash),
            'produced_at': produced_at,
            'responses': responses,
            'response_extensions': response_data_extensions
        })

        signature_algo = responder_private_key.algorithm
        if signature_algo == 'ec':
            signature_algo = 'ecdsa'

        signature_algorithm_id = '%s_%s' % (self._hash_algo, signature_algo)

        if responder_private_key.algorithm == 'rsa':
            sign_func = asymmetric.rsa_pkcs1v15_sign
        elif responder_private_key.algorithm == 'dsa':
            sign_func = asymmetric.dsa_sign
        elif responder_private_key.algorithm == 'ec':
            sign_func = asymmetric.ecdsa_sign

        if not is_oscrypto:
            responder_private_key = asymmetric.load_private_key(responder_private_key)
        signature_bytes = sign_func(responder_private_key, response_data.dump(), self._hash_algo)

        certs = None
        if self._certificate_issuer and getattr(self._certificate_issuer.public_key, self._key_hash_algo) != responder_key_hash:
            certs = [responder_certificate]

        return ocsp.OCSPResponse({
            'response_status': self._response_status,
            'response_bytes': {
                'response_type': 'basic_ocsp_response',
                'response': {
                    'tbs_response_data': response_data,
                    'signature_algorithm': {'algorithm': signature_algorithm_id},
                    'signature': signature_bytes,
                    'certs': certs,
                }
            }
        })

# Enums

class ResponseStatus(enum.Enum):
    successful = 'successful'
    malformed_request = 'malformed_request'
    internal_error = 'internal_error'
    try_later = 'try_later'
    sign_required = 'sign_required'
    unauthorized = 'unauthorized'


class CertificateStatus(enum.Enum):
    good = 'good'
    revoked = 'revoked'
    key_compromise = 'key_compromise'
    ca_compromise = 'ca_compromise'
    affiliation_changed = 'affiliation_changed'
    superseded = 'superseded'
    cessation_of_operation = 'cessation_of_operation'
    certificate_hold = 'certificate_hold'
    remove_from_crl = 'remove_from_crl'
    privilege_withdrawn = 'privilege_withdrawn'
    unknown = 'unknown'


# API endpoints
FAULT_REVOKED = "revoked"
FAULT_UNKNOWN = "unknown"

app = Flask(__name__)
class OCSPResponder:

    def __init__(self, issuer_cert: str, responder_cert: str, responder_key: str,
                       fault: str, next_update_seconds: int):
        """
        Create a new OCSPResponder instance.

        :param issuer_cert: Path to the issuer certificate.
        :param responder_cert: Path to the certificate of the OCSP responder
            with the `OCSP Signing` extension.
        :param responder_key: Path to the private key belonging to the
            responder cert.
        :param validate_func: A function that - given a certificate serial -
            will return the appropriate :class:`CertificateStatus` and -
            depending on the status - a revocation datetime.
        :param cert_retrieve_func: A function that - given a certificate serial -
            will return the corresponding certificate as a string.
        :param next_update_seconds: The ``nextUpdate`` value that will be written
            into the response. Default: 9 hours.

        """
        # Certs and keys
        self._issuer_cert = asymmetric.load_certificate(issuer_cert)
        self._responder_cert = asymmetric.load_certificate(responder_cert)
        self._responder_key = asymmetric.load_private_key(responder_key)

        # Next update
        self._next_update_seconds = next_update_seconds

        self._fault = fault

    def _fail(self, status: ResponseStatus) -> OCSPResponse:
        builder = OCSPResponseBuilder(response_status=status.value)
        return builder.build()

    def parse_ocsp_request(self, request_der: bytes) -> OCSPRequest:
        """
        Parse the request bytes, return an ``OCSPRequest`` instance.
        """
        return OCSPRequest.load(request_der)

    def validate(self):
        time = datetime(2018, 1, 1, 1, 00, 00, 00, timezone.utc)
        if self._fault == FAULT_REVOKED:
            return (CertificateStatus.revoked, time)
        elif self._fault == FAULT_UNKNOWN:
            return (CertificateStatus.unknown, None)
        elif self._fault != None:
            raise NotImplemented('Fault type could not be found')
        return (CertificateStatus.good, time)

    def _build_ocsp_response(self, ocsp_request: OCSPRequest) -> OCSPResponse:
        """
        Create and return an OCSP response from an OCSP request.
        """
        # Get the certificate serial
        tbs_request = ocsp_request['tbs_request']
        request_list = tbs_request['request_list']
        if len(request_list) < 1:
            logger.warning('Received OCSP request with no requests')
            raise NotImplemented('Empty requests not supported')

        single_request = request_list[0]  # TODO: Support more than one request
        req_cert = single_request['req_cert']
        serial = req_cert['serial_number'].native

        # Check certificate status
        try:
            certificate_status, revocation_date = self.validate()
        except Exception as e:
            logger.exception('Could not determine certificate status: %s', e)
            return self._fail(ResponseStatus.internal_error)

        certificate_status_list = [(serial, certificate_status.value)]

        # Build the response
        builder = OCSPResponseBuilder(**{
            'response_status': ResponseStatus.successful.value,
            'certificate_status_list': certificate_status_list,
            'revocation_date': revocation_date,
        })

        # Parse extensions
        for extension in tbs_request['request_extensions']:
            extn_id = extension['extn_id'].native
            critical = extension['critical'].native
            value = extension['extn_value'].parsed

            # This variable tracks whether any unknown extensions were encountered
            unknown = False

            # Handle nonce extension
            if extn_id == 'nonce':
                builder.nonce = value.native

            # That's all we know
            else:
                unknown = True

            # If an unknown critical extension is encountered (which should not
            # usually happen, according to RFC 6960 4.1.2), we should throw our
            # hands up in despair and run.
            if unknown is True and critical is True:
                logger.warning('Could not parse unknown critical extension: %r',
                        dict(extension.native))
                return self._fail(ResponseStatus.internal_error)

            # If it's an unknown non-critical extension, we can safely ignore it.
            elif unknown is True:
                logger.info('Ignored unknown non-critical extension: %r', dict(extension.native))

        # Set certificate issuer
        builder.certificate_issuer = self._issuer_cert

        # Set next update date
        now = datetime.now(timezone.utc)
        builder.next_update = (now + timedelta(seconds=self._next_update_seconds)).replace(microsecond=0)

        return builder.build(self._responder_key, self._responder_cert)

    def build_http_response(self, request_der: bytes) -> Response:
        global app
        response_der = self._build_ocsp_response(request_der).dump()
        resp = app.make_response((response_der, 200))
        resp.headers['content_type'] = 'application/ocsp-response'
        return resp


responder = None

def init_responder(issuer_cert: str, responder_cert: str, responder_key: str, fault: str, next_update_seconds: int):
    global responder
    responder = OCSPResponder(issuer_cert=issuer_cert, responder_cert=responder_cert, responder_key=responder_key, fault=fault, next_update_seconds=next_update_seconds)

def init(port=8080, debug=False):
    logger.info('Launching %sserver on port %d', 'debug' if debug else '', port)
    app.run(port=port, debug=debug)

@app.route('/', methods=['GET'])
def _handle_root():
    return 'ocsp-responder'

@app.route('/status/', defaults={'u_path': ''}, methods=['GET'])
@app.route('/status/<path:u_path>', methods=['GET'])
def _handle_get(u_path):
    global responder
    """
    An OCSP GET request contains the DER-in-base64 encoded OCSP request in the
    HTTP request URL.
    """
    if "Host" not in request.headers:
        raise ValueError ("Required 'Host' header not present")
    der = base64.b64decode(u_path)
    ocsp_request = responder.parse_ocsp_request(der)
    return responder.build_http_response(ocsp_request)

@app.route('/status', methods=['POST'])
def _handle_post():
    global responder
    """
    An OCSP POST request contains the DER encoded OCSP request in the HTTP
    request body.
    """
    if "Host" not in request.headers:
        raise ValueError ("Required 'Host' header not present")
    ocsp_request = responder.parse_ocsp_request(request.data)
    return responder.build_http_response(ocsp_request)



# Copyright 2009-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict as OD

from evergreen_config_generator.functions import (
    Function, s3_put, shell_exec)
from evergreen_config_lib import shell_mongoc

build_path = '${build_variant}/${revision}/${version_id}/${build_id}'

all_functions = OD([
    ('install ssl', Function(
        shell_mongoc(r'''
        .evergreen/scripts/install-ssl.sh
        ''', test=False, add_expansions_to_env=True),
    )),
    ('upload coverage', Function(
        shell_mongoc(r'''
        export AWS_ACCESS_KEY_ID=${aws_key}
        export AWS_SECRET_ACCESS_KEY=${aws_secret}
        aws s3 cp coverage s3://mciuploads/mongo-c-driver/%s/coverage/ --recursive --acl public-read --region us-east-1
        ''' % (build_path,), test=False, silent=True),
        s3_put(build_path + '/coverage/index.html', aws_key='${aws_key}',
               aws_secret='${aws_secret}',
               local_file='mongoc/coverage/index.html', bucket='mciuploads',
               permissions='public-read', content_type='text/html',
               display_name='Coverage Report'),
    )),
    ('upload scan artifacts', Function(
        shell_mongoc(r'''
        if find scan -name \*.html | grep -q html; then
          (cd scan && find . -name index.html -exec echo "<li><a href='{}'>{}</a></li>" \;) >> scan.html
        else
          echo "No issues found" > scan.html
        fi
        '''),
        shell_mongoc(r'''
        export AWS_ACCESS_KEY_ID=${aws_key}
        export AWS_SECRET_ACCESS_KEY=${aws_secret}
        aws s3 cp scan s3://mciuploads/mongo-c-driver/%s/scan/ --recursive --acl public-read --region us-east-1
        ''' % (build_path,), test=False, silent=True),
        s3_put(build_path + '/scan/index.html', aws_key='${aws_key}',
               aws_secret='${aws_secret}', local_file='mongoc/scan.html',
               bucket='mciuploads', permissions='public-read',
               content_type='text/html', display_name='Scan Build Report'),
    )),
    # Use "silent=True" to hide output since errors may contain credentials.
    ('run auth tests', Function(
        shell_mongoc(r'''
        .evergreen/scripts/run-auth-tests.sh
        ''', add_expansions_to_env=True),
    )),
    ('link sample program', Function(
        shell_mongoc(r'''
        # Compile a program that links dynamically or statically to libmongoc,
        # using variables from pkg-config or CMake's find_package command.
        export BUILD_SAMPLE_WITH_CMAKE=${BUILD_SAMPLE_WITH_CMAKE}
        export BUILD_SAMPLE_WITH_CMAKE_DEPRECATED=${BUILD_SAMPLE_WITH_CMAKE_DEPRECATED}
        export ENABLE_SSL=${ENABLE_SSL}
        export ENABLE_SNAPPY=${ENABLE_SNAPPY}
        LINK_STATIC=  .evergreen/scripts/link-sample-program.sh
        LINK_STATIC=1 .evergreen/scripts/link-sample-program.sh
        ''',
        include_expansions_in_env=['distro_id']),
    )),
    ('link sample program bson', Function(
        shell_mongoc(r'''
        # Compile a program that links dynamically or statically to libbson,
        # using variables from pkg-config or from CMake's find_package command.
        BUILD_SAMPLE_WITH_CMAKE=  BUILD_SAMPLE_WITH_CMAKE_DEPRECATED=  LINK_STATIC=  .evergreen/scripts/link-sample-program-bson.sh
        BUILD_SAMPLE_WITH_CMAKE=  BUILD_SAMPLE_WITH_CMAKE_DEPRECATED=  LINK_STATIC=1 .evergreen/scripts/link-sample-program-bson.sh
        BUILD_SAMPLE_WITH_CMAKE=1 BUILD_SAMPLE_WITH_CMAKE_DEPRECATED=  LINK_STATIC=  .evergreen/scripts/link-sample-program-bson.sh
        BUILD_SAMPLE_WITH_CMAKE=1 BUILD_SAMPLE_WITH_CMAKE_DEPRECATED=  LINK_STATIC=1 .evergreen/scripts/link-sample-program-bson.sh
        BUILD_SAMPLE_WITH_CMAKE=1 BUILD_SAMPLE_WITH_CMAKE_DEPRECATED=1 LINK_STATIC=  .evergreen/scripts/link-sample-program-bson.sh
        BUILD_SAMPLE_WITH_CMAKE=1 BUILD_SAMPLE_WITH_CMAKE_DEPRECATED=1 LINK_STATIC=1 .evergreen/scripts/link-sample-program-bson.sh
        ''',
        include_expansions_in_env=['distro_id']),
    )),
    ('link sample program MSVC', Function(
        shell_mongoc(r'''
        # Build libmongoc with CMake and compile a program that links
        # dynamically or statically to it, using variables from CMake's
        # find_package command.
        export ENABLE_SSL=${ENABLE_SSL}
        export ENABLE_SNAPPY=${ENABLE_SNAPPY}
        . .evergreen/scripts/use-tools.sh paths
        . .evergreen/scripts/find-cmake-latest.sh
        export CMAKE="$(native-path "$(find_cmake_latest)")"
        LINK_STATIC=  cmd.exe /c .\\.evergreen\\scripts\\link-sample-program-msvc.cmd
        LINK_STATIC=1 cmd.exe /c .\\.evergreen\\scripts\\link-sample-program-msvc.cmd
        ''',
        include_expansions_in_env=['distro_id']),
    )),
    ('link sample program mingw', Function(
        shell_mongoc(r'''
        # Build libmongoc with CMake and compile a program that links
        # dynamically to it, using variables from pkg-config.exe.
        . .evergreen/scripts/use-tools.sh paths
        . .evergreen/scripts/find-cmake-latest.sh
        export CMAKE="$(native-path "$(find_cmake_latest)")"
        cmd.exe /c .\\.evergreen\\scripts\\link-sample-program-mingw.cmd
        ''',
        include_expansions_in_env=['distro_id']),
    )),
    ('link sample program MSVC bson', Function(
        shell_mongoc(r'''
        # Build libmongoc with CMake and compile a program that links
        # dynamically or statically to it, using variables from CMake's
        # find_package command.
        export ENABLE_SSL=${ENABLE_SSL}
        export ENABLE_SNAPPY=${ENABLE_SNAPPY}
        . .evergreen/scripts/use-tools.sh paths
        . .evergreen/scripts/find-cmake-latest.sh
        export CMAKE="$(native-path "$(find_cmake_latest)")"
        LINK_STATIC=  cmd.exe /c .\\.evergreen\\scripts\\link-sample-program-msvc-bson.cmd
        LINK_STATIC=1 cmd.exe /c .\\.evergreen\\scripts\\link-sample-program-msvc-bson.cmd
        ''',
        include_expansions_in_env=['distro_id']),
    )),
    ('link sample program mingw bson', Function(
        shell_mongoc(r'''
        # Build libmongoc with CMake and compile a program that links
        # dynamically to it, using variables from pkg-config.exe.
        . .evergreen/scripts/use-tools.sh paths
        . .evergreen/scripts/find-cmake-latest.sh
        export CMAKE="$(native-path "$(find_cmake_latest)")"
        cmd.exe /c .\\.evergreen\\scripts\\link-sample-program-mingw-bson.cmd
        '''),
    )),
    ('update codecov.io', Function(
        shell_mongoc(r'''
        # Note: coverage is currently only enabled on the ubuntu1804 distro.
        # This script does not support MacOS, Windows, or non-x86_64 distros.
        # Update accordingly if code coverage is expanded to other distros.
        curl -Os https://uploader.codecov.io/latest/linux/codecov
        chmod +x codecov
        # -Z: Exit with a non-zero value if error.
        # -g: Run with gcov support.
        # -t: Codecov upload token.
        # perl: filter verbose "Found" list and "Processing" messages.
        ./codecov -Zgt "${codecov_token}" | perl -lne 'print if not m|^.*\.gcov(\.\.\.)?$|'
        ''', test=False),
    )),
    ('compile coverage', Function(
        shell_mongoc(r'''
        COVERAGE=ON .evergreen/scripts/compile.sh
        ''', add_expansions_to_env=True),
    )),
    ('build mongohouse', Function(
        shell_exec(r'''
        cd drivers-evergreen-tools
        export DRIVERS_TOOLS=$(pwd)
        .evergreen/atlas_data_lake/pull-mongohouse-image.sh
        '''),
    )),
    ('run mongohouse', Function(
        shell_exec(r'''
        cd drivers-evergreen-tools
        export DRIVERS_TOOLS=$(pwd)
        .evergreen/atlas_data_lake/run-mongohouse-image.sh
        '''),
    )),
    ('test mongohouse', Function(
        shell_mongoc(r'''
        echo "Waiting for mongohouse to start..."
        wait_for_mongohouse() {
            for _ in $(seq 300); do
                # Exit code 7: "Failed to connect to host".
                if curl -s localhost:$1; (("$?" != 7)); then
                    return 0
                else
                    sleep 1
                fi
            done
            echo "Could not detect mongohouse on port $1" 1>&2
            return 1
        }
        wait_for_mongohouse 27017 || exit
        echo "Waiting for mongohouse to start... done."
        pgrep -a "mongohouse"
        export RUN_MONGOHOUSE_TESTS=ON
        ./cmake-build/src/libmongoc/test-libmongoc --no-fork -l /mongohouse/* -d --skip-tests .evergreen/etc/skip-tests.txt
        '''),
    )),
    ('run aws tests', Function(
        # Assume role to get AWS secrets.
        {
            "command": "ec2.assume_role",
            "params": {
                "role_arn": "${aws_test_secrets_role}"
            }
        },

        shell_mongoc(r'''
        pushd ../drivers-evergreen-tools/.evergreen/auth_aws
        ./setup_secrets.sh drivers/aws_auth
        popd # ../drivers-evergreen-tools/.evergreen/auth_aws
        ''', include_expansions_in_env=["AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN"]),

        shell_mongoc(r'''
        pushd ../drivers-evergreen-tools/.evergreen/auth_aws
        . ./activate-authawsvenv.sh
        popd # ../drivers-evergreen-tools/.evergreen/auth_aws
        .evergreen/scripts/run-aws-tests.sh
        ''', include_expansions_in_env=["TESTCASE"])
    )),
])



"""
This is a miniature Sphinx extension that is only used to permit resolution of
external references to the CMake Sphinx documentation. It defines no additional
directives or object types.

This extension defines a "cmake" domain, which may conflict with a proper CMake
domain plugin. If that is the case, this extention can likely be disabled and
replaced by a more full-featured extension.

"""
from typing import Any, List
from sphinx.application import Sphinx
from sphinx.roles import XRefRole
from sphinx.domains import Domain, ObjType

kinds = [
    "command",
    "cpack_gen",
    "envvar",
    "generator",
    "genex",
    "guide",
    "variable",
    "module",
    "policy",
    "prop_cache",
    "prop_dir",
    "prop_gbl",
    "prop_inst",
    "prop_sf",
    "prop_test",
    "prop_tgt",
    "manual",
]


class CMakeRefDomain(Domain):
    name = "cmake"
    label = "CMake (Minimal)"
    object_types = {k: ObjType(k, k) for k in kinds}
    roles = {k: XRefRole() for k in kinds}
    roles["command"] = XRefRole(fix_parens=True)
    directives = {}
    initial_data: Any = {}

    def merge_domaindata(self, docnames: List[str], otherdata: Any) -> None:
        # We have nothing to do, but this is required for parallel execution
        return


def setup(app: Sphinx):
    app.add_domain(CMakeRefDomain)
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



#!/usr/bin/env python3

# Generates *.yml files under .evergreen/generated_configs.


import sys

from importlib import import_module


GENERATOR_NAMES = [
    "functions",
    "tasks",
    "task_groups",
    "variants",
    "legacy_config",
]


def main():
    # Requires Python 3.10 or newer.
    assert sys.version_info.major >= 3
    assert sys.version_info.minor >= 10

    for name in GENERATOR_NAMES:
        m = import_module(f"config_generator.generators.{name}")
        print(f"Running {name}.generate()...")
        m.generate()
        print(f"Running {name}.generate()... done.")


if __name__ == "__main__":
    main()



# Copyright 2009-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict as OD
from textwrap import dedent
from typing import Iterable

from evergreen_config_generator import ConfigObject

from . import Value, MutableValueMapping, ValueMapping, ValueOrderedDict


def func(func_name: str, **kwargs: Value) -> MutableValueMapping:
    od: MutableValueMapping = OD([("func", func_name)])
    if kwargs:
        od["vars"] = OD(sorted(kwargs.items()))

    return od


def s3_put(remote_file: str, project_path: bool = True, **kwargs: Value) -> ValueMapping:
    if project_path:
        remote_file = "mongo-c-driver/" + remote_file

    return ValueOrderedDict(
        [
            ("command", "s3.put"),
            (
                "params",
                ValueOrderedDict(
                    (
                        ("aws_key", "${aws_key}"),
                        ("aws_secret", "${aws_secret}"),
                        ("remote_file", remote_file),
                        ("bucket", "mciuploads"),
                        ("permissions", "public-read"),
                        *kwargs.items(),
                    )
                ),
            ),
        ]
    )


def strip_lines(s: str) -> str:
    return "\n".join(line for line in s.split("\n") if line.strip())


def shell_exec(
    script: str,
    test: bool = True,
    errexit: bool = True,
    xtrace: bool = False,
    silent: bool = False,
    continue_on_err: bool = False,
    working_dir: str | None = None,
    background: bool = False,
    add_expansions_to_env: bool = False,
    redirect_standard_error_to_output: bool = False,
    include_expansions_in_env: Iterable[str] = (),
) -> ValueMapping:
    dedented = ""
    if errexit:
        dedented += "set -o errexit\n"

    if xtrace:
        dedented += "set -o xtrace\n"

    dedented += dedent(strip_lines(script))
    command = ValueOrderedDict([("command", "shell.exec")])
    if test:
        command["type"] = "test"

    command["params"] = OD()
    if silent:
        command["params"]["silent"] = True

    if working_dir is not None:
        command["params"]["working_dir"] = working_dir

    if continue_on_err:
        command["params"]["continue_on_err"] = True

    if background:
        command["params"]["background"] = True

    if add_expansions_to_env:
        command["params"]["add_expansions_to_env"] = True

    if redirect_standard_error_to_output:
        command["params"]["redirect_standard_error_to_output"] = True

    if include_expansions_in_env:
        command["params"]["include_expansions_in_env"] = list(include_expansions_in_env)

    command["params"]["shell"] = "bash"
    command["params"]["script"] = dedented
    return command


def targz_pack(target: str, source_dir: str, *include: str) -> ValueMapping:
    return OD(
        [
            ("command", "archive.targz_pack"),
            ("params", OD([("target", target), ("source_dir", source_dir), ("include", list(include))])),
        ]
    )


class Function(ConfigObject):
    def __init__(self, *commands: Value):
        super(Function, self).__init__()
        self.commands = commands

    def to_dict(self) -> Value:
        return list(self.commands)



from shrub.v3.evg_command import EvgCommandType

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class RunTests(Function):
    name = 'run-tests'
    commands = [
        bash_exec(
            command_type=EvgCommandType.TEST,
            script='.evergreen/scripts/run-tests.sh',
            working_dir='mongoc',
            add_expansions_to_env=True,
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return RunTests.defn()



"""
This Python type stubs file is only here to support checking of lldb_bson.py.
The type/member/constant listings in here are incomplete, as only those that
are used in lldb_bson have been transcribed from the LLDB Python API
documentation. Refer: https://lldb.llvm.org/python_api.html
"""

from typing import IO, Any, Sequence, TypeAlias, NoReturn

_Pointer: TypeAlias = int
_Size: TypeAlias = int
_Offset: TypeAlias = int
_InternalDict: TypeAlias = dict[str, Any]

class SBError:
    def __init__(self) -> None: ...
    @property
    def success(self) -> bool: ...
    @property
    def fail(self) -> bool: ...
    @property
    def description(self) -> str: ...

class SBDebugger:
    def __init__(self, _: NoReturn) -> None: ...
    def HandleCommand(self, command: str) -> None: ...
    def GetErrorFileHandle(self) -> IO[str]: ...
    def GetSelectedTarget(self) -> SBTarget | None: ...

class SBAddress:
    def __init__(self) -> None: ...
    @property
    def offset(self) -> type: ...
    @property
    def load_addr(self) -> _Pointer: ...
    def SetLoadAddress(self, addr: _Pointer, tgt: SBTarget) -> None: ...

class SBData:
    def GetSignedInt32(self, err: SBError, off: _Offset) -> int: ...
    @staticmethod
    def CreateDataFromUInt64Array(endian: ByteOrderType, size: _Size, dat: Sequence[int]) -> SBData: ...
    @staticmethod
    def CreateDataFromCString(endian: ByteOrderType, size: _Size, s: str) -> SBData: ...

    byte_order: ByteOrderType

    @property
    def uint8(self) -> Sequence[int]: ...
    @property
    def uint32(self) -> Sequence[int]: ...
    @property
    def uint64(self) -> Sequence[int]: ...
    @property
    def size(self) -> int: ...
    def ReadRawData(self, err: SBError, off: _Offset, size: _Size) -> bytes: ...
    def GetAddress(self, err: SBError, off: _Offset) -> _Pointer: ...
    def GetUnsignedInt32(self, err: SBError, off: _Offset) -> int: ...
    def GetUnsignedInt64(self, err: SBError, off: _Offset) -> int: ...

class SBValue:
    def Dereference(self) -> SBValue: ...
    def TypeIsPointerType(self) -> bool: ...
    def GetChildMemberWithName(self, name: str) -> SBValue: ...
    def GetValueAsUnsigned(self) -> int: ...
    def Cast(self, type: SBType) -> SBValue: ...
    def GetAddress(self) -> SBAddress: ...
    def GetData(self) -> SBData: ...
    def GetNonSyntheticValue(self) -> SBValue: ...
    def CreateChildAtOffset(self, name: str, offset: _Offset, type: SBType) -> SBValue: ...
    def CreateValueFromData(self, name: str, data: SBData, type: SBType) -> SBValue: ...
    def CreateValueFromAddress(self, name: str, addr: _Pointer, type: SBType) -> SBValue: ...
    def CreateValueFromExpression(self, name: str, expr: str) -> SBValue: ...
    def synthetic_child_from_address(self, name: str, addr: _Pointer, type: SBType) -> SBValue: ...
    def synthetic_child_from_expression(self, name: str, expr: str) -> SBValue: ...
    @property
    def addr(self) -> SBAddress: ...
    @property
    def frame(self) -> SBFrame: ...
    @property
    def type(self) -> SBType: ...
    @property
    def name(self) -> str: ...
    @property
    def target(self) -> SBTarget: ...
    @property
    def load_addr(self) -> _Pointer: ...
    @property
    def size(self) -> int: ...
    @property
    def unsigned(self) -> int: ...
    @property
    def data(self) -> SBData: ...
    @property
    def deref(self) -> SBValue: ...
    @property
    def process(self) -> SBProcess: ...
    @property
    def error(self) -> SBError: ...
    @property
    def changed(self) -> bool: ...
    @property
    def children(self) -> Sequence[SBValue]: ...

    value: str
    format: ValueFormatType

class SBFrame:
    def FindVariable(self, var_name: str) -> SBValue: ...
    def EvaluateExpression(self, expr: str) -> SBValue: ...
    @property
    def locals(self) -> Sequence[SBValue]: ...
    @property
    def thread(self) -> SBThread: ...

class SBThread:
    @property
    def frames(self) -> Sequence[SBFrame]: ...
    @property
    def process(self) -> SBProcess: ...

class SBProcess:
    def ReadMemory(self, addr: _Pointer, size: int, err: SBError) -> bytes: ...
    def ReadCStringFromMemory(self, addr: _Pointer, max: int, err: SBError) -> bytes: ...
    @property
    def selected_thread(self) -> SBThread: ...
    @property
    def id(self) -> int: ...

class SBCommandReturnObject:
    def AppendMessage(self, s: str) -> None: ...

class SBTarget:
    def FindFirstType(self, type: str) -> SBType: ...
    def CreateValueFromData(self, name: str, data: SBData, type: SBType) -> SBValue: ...
    def CreateValueFromExpression(self, name: str, expr: str) -> SBValue: ...
    def CreateValueFromAddress(self, name: str, addr: SBAddress, type: SBType) -> SBValue: ...
    def GetBasicType(self, type: BasicType) -> SBType: ...
    @property
    def process(self) -> SBProcess: ...

class SBType:
    def GetDisplayTypeName(self) -> str: ...
    def GetPointerType(self) -> SBType: ...
    def GetArrayType(self, size: int) -> SBType: ...
    @property
    def name(self) -> str: ...
    def IsValid(self) -> bool: ...
    @property
    def fields(self) -> Sequence[SBTypeMember]: ...
    @property
    def size(self) -> int: ...

class SBTypeMember:
    @property
    def name(self) -> str: ...
    @property
    def byte_offset(self) -> int: ...

class SBSyntheticValueProvider:
    def __init__(self, valobj: SBValue, internal_dict: _InternalDict) -> None: ...
    def num_children(self) -> int: ...
    def get_child_at_index(self, pos: int) -> SBValue: ...
    def has_children(self) -> bool: ...
    def get_value(self) -> SBValue: ...
    def update(self) -> bool | None: ...

# Types not present in lldb, but represent the types of top-level constants:
# (These may only be used in unevaluated contexts)
class ByteOrderType: ...
class ValueFormatType: ...
class BasicType: ...

eBasicTypeNullPtr: BasicType
eBasicTypeUnsignedChar: BasicType
eBasicTypeDouble: BasicType
eBasicTypeBool: BasicType
eBasicTypeChar: BasicType
eBasicTypeUnsignedChar: BasicType
eBasicTypeVoid: BasicType

eFormatDefault: ValueFormatType
eFormatHex: ValueFormatType
eFormatBinary: ValueFormatType
eFormatPointer: ValueFormatType
eFormatInvalid: ValueFormatType

debugger: SBDebugger



from shrub.v3.evg_project import EvgProject

from config_generator.etc import utils


def generate():
    task_groups = []

    for component in utils.all_components():
        if hasattr(component, 'task_groups'):
            component_name = utils.component_name(component)
            print(f' - {component_name}')
            task_groups += component.task_groups()

    task_groups.sort(key=lambda v: v.name)
    yaml = utils.to_yaml(EvgProject(task_groups=task_groups))
    utils.write_to_file(yaml, 'task_groups.yml')



from shrub.v3.evg_command import EvgCommandType
from shrub.v3.evg_command import s3_put
from shrub.v3.evg_task import EvgTask

from config_generator.components.funcs.set_cache_dir import SetCacheDir

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class CheckABICompliance(Function):
    name = 'abi-compliance-check'
    commands = SetCacheDir.commands + [
        bash_exec(
            command_type=EvgCommandType.SETUP,
            working_dir='mongoc',
            include_expansions_in_env=['MONGO_C_DRIVER_CACHE_DIR'],
            script='.evergreen/scripts/abi-compliance-check-setup.sh'
        ),
        bash_exec(
            command_type=EvgCommandType.TEST,
            add_expansions_to_env=True,
            working_dir='mongoc',
            include_expansions_in_env=['MONGO_C_DRIVER_CACHE_DIR'],
            script='.evergreen/scripts/abi-compliance-check.sh'
        ),
        s3_put(
            command_type=EvgCommandType.SYSTEM,
            aws_key='${aws_key}',
            aws_secret='${aws_secret}',
            bucket='mciuploads',
            content_type='text/html',
            display_name='ABI Report:',
            local_files_include_filter='mongoc/abi-compliance/compat_reports/**/*.html',
            permissions='public-read',
            remote_file='mongo-c-driver/${branch_name}/${revision}/${version_id}/${build_id}/${task_id}/${execution}/abi-compliance-check/',
        ),
        s3_put(
            command_type=EvgCommandType.SYSTEM,
            aws_key='${aws_key}',
            aws_secret='${aws_secret}',
            bucket='mciuploads',
            content_type='text/plain',
            display_name='ABI Compliance Check: ',
            local_files_include_filter='abi-compliance/logs/**/log.txt',
            permissions='public-read',
            remote_file='mongo-c-driver/${branch_name}/${revision}/${version_id}/${build_id}/${task_id}/${execution}/abi-compliance-check/',
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return CheckABICompliance.defn()


def tasks():
    return [
        EvgTask(
            name=CheckABICompliance.name,
            commands=[CheckABICompliance.call()],
        )
    ]



from shrub.v3.evg_build_variant import BuildVariant
from shrub.v3.evg_task import EvgTaskRef

from config_generator.etc.compile import generate_compile_tasks

from config_generator.etc.cse.compile import CompileCommon
from config_generator.etc.cse.test import generate_test_tasks


SSL = 'winssl'
TAG = f'cse-matrix-{SSL}'


# pylint: disable=line-too-long
# fmt: off
COMPILE_MATRIX = [
    ('windows-vsCurrent', 'vs2017x64', None, ['cyrus']),
]

# TODO (CDRIVER-3789): test cse with the 'sharded' topology.
TEST_MATRIX = [
    ('windows-vsCurrent', 'vs2017x64', None, 'cyrus', ['auth'], ['server'], ['4.2', '4.4', '5.0', '6.0'                ]),

    # Test 7.0+ with a replica set since Queryable Encryption does not support the 'server' topology. Queryable Encryption tests require 7.0+.
    ('windows-vsCurrent', 'vs2017x64', None, 'cyrus', ['auth'], ['server', 'replica' ], [               '7.0', '8.0', 'latest']),
]
# fmt: on
# pylint: enable=line-too-long


class WinSSLCompileCommon(CompileCommon):
    ssl = 'WINDOWS'


class SaslCyrusWinSSLCompile(WinSSLCompileCommon):
    name = 'cse-sasl-cyrus-winssl-compile'
    commands = WinSSLCompileCommon.compile_commands(sasl='CYRUS')


def functions():
    return SaslCyrusWinSSLCompile.defn()


def tasks():
    res = []

    SASL_TO_FUNC = {
        'cyrus': SaslCyrusWinSSLCompile,
    }

    MORE_TAGS = ['cse']

    res += generate_compile_tasks(
        SSL, TAG, SASL_TO_FUNC, COMPILE_MATRIX, MORE_TAGS
    )

    res += generate_test_tasks(SSL, TAG, TEST_MATRIX)

    return res


def variants():
    expansions = {
        'CLIENT_SIDE_ENCRYPTION': 'on',
    }

    return [
        BuildVariant(
            name=TAG,
            display_name=TAG,
            tasks=[EvgTaskRef(name=f'.{TAG}')],
            expansions=expansions,
        ),
    ]



from typing import List

from shrub.v3.evg_build_variant import BuildVariant

from config_generator.etc.utils import TaskRef
from config_generator.etc.function import merge_defns
from config_generator.etc.compile import generate_compile_tasks

from config_generator.etc.sasl.compile import CompileCommon
from config_generator.etc.sasl.test import generate_test_tasks


SSL = 'openssl'
TAG = f'sasl-matrix-{SSL}'


# pylint: disable=line-too-long
# fmt: off
COMPILE_MATRIX = [
    ('debian92',          'clang',      None, ['cyrus']),
    ('debian92',          'gcc',        None, ['cyrus']),
    ('debian10',          'gcc',        None, ['cyrus']),
    ('debian11',          'gcc',        None, ['cyrus']),
    ('rhel80',            'gcc',        None, ['cyrus']),
    ('rhel8-power',       'gcc',        None, ['cyrus']),
    ('rhel8-zseries',     'gcc',        None, ['cyrus']),
    ('ubuntu2004',        'clang',      None, ['cyrus']),
    ('ubuntu2004-arm64',  'gcc',        None, ['cyrus']),
    ('ubuntu2004',        'gcc',        None, ['cyrus']),
    ('windows-vsCurrent', 'vs2017x64',  None, ['cyrus']),
]

TEST_MATRIX = [
    ('rhel8-power',  'gcc', None, 'cyrus', ['auth'], ['server'], ['4.2', '4.4', '5.0', '6.0', '7.0', '8.0', 'latest']),
    ('rhel8-zseries', 'gcc', None, 'cyrus', ['auth'], ['server'], [              '5.0', '6.0', '7.0', '8.0', 'latest']),

    ('ubuntu2004-arm64',  'gcc',       None, 'cyrus', ['auth'], ['server'], ['4.4', '5.0', '6.0', '7.0', '8.0', 'latest']),
    ('ubuntu2004',        'gcc',       None, 'cyrus', ['auth'], ['server'], ['4.4', '5.0', '6.0', '7.0', '8.0', 'latest']),
    ('windows-vsCurrent', 'vs2017x64', None, 'cyrus', ['auth'], ['server'], [       'latest']),

    # Test 4.2 with Debian 10 since 4.2 does not ship on Ubuntu 20.04+.
    ('debian10',          'gcc',       None, 'cyrus', ['auth'], ['server', 'replica'], ['4.2']), 
]
# fmt: on
# pylint: enable=line-too-long


class OpenSSLCompileCommon(CompileCommon):
    ssl = 'OPENSSL'


class SaslCyrusOpenSSLCompile(OpenSSLCompileCommon):
    name = 'sasl-cyrus-openssl-compile'
    commands = OpenSSLCompileCommon.compile_commands(sasl='CYRUS')


def functions():
    return merge_defns(
        SaslCyrusOpenSSLCompile.defn(),
    )


SASL_TO_FUNC = {
    'cyrus': SaslCyrusOpenSSLCompile,
}

TASKS = [
    *generate_compile_tasks(SSL, TAG, SASL_TO_FUNC, COMPILE_MATRIX),
    *generate_test_tasks(SSL, TAG, TEST_MATRIX),
]


def tasks():
    res = TASKS.copy()

    # PowerPC and zSeries are limited resources.
    for task in res:
        if any(pattern in task.run_on for pattern in ["power", "zseries"]):
            task.patchable = False

    return res


def variants():
    expansions = {}

    tasks = []

    # PowerPC and zSeries are limited resources.
    for task in TASKS:
        if any(pattern in task.run_on for pattern in ["power", "zseries"]):
            tasks.append(
                TaskRef(
                    name=task.name,
                    batchtime=1440,   # 1 day
                )
            )
        else:
            tasks.append(task.get_task_ref())

    return [
        BuildVariant(
            name=TAG,
            display_name=TAG,
            tasks=tasks,
            expansions=expansions,
        ),
    ]



from typing import Literal

from pydantic import BaseModel, validator
from packaging.version import Version


class Distro(BaseModel):
    """
    Defines common properties of a given Evergreen distro.

    * name: Name of the distro.
    * os: Name of the operating system.
    * os_type: One of Linux, MacOS, or Windows.
    * os_ver: Version of the operating system.
    * vs_ver: Version of Visual Studio available.
    * size: Size of tasks the distro is designed to handle.
    * arch: Target architecture.
    """

    name: str
    os: str | None = None
    os_type: Literal['linux', 'macos', 'windows'] | None = None
    os_ver: str | None = None
    vs_ver: Literal[
        '2013',
        '2015',
        '2017',
        '2019',
        '2022',
        'vsCurrent',
        'vsCurrent2',
        'vsMulti',
    ] | None = None
    size: Literal['small', 'large'] | None = None
    arch: Literal['arm64', 'power', 'zseries'] | None = None

    @validator('os_ver')
    @classmethod
    def validate_os_ver(cls, value):
        return Version(value)

# See: https://evergreen.mongodb.com/distros
# pylint: disable=line-too-long
#fmt: off
DEBIAN_DISTROS = [
    Distro(name='debian92-large', os='debian', os_type='linux', os_ver='9.2', size='large'), # CDRIVER-5873
    Distro(name='debian92-small', os='debian', os_type='linux', os_ver='9.2', size='small'), # CDRIVER-5873
    Distro(name='debian10-large', os='debian', os_type='linux', os_ver='10', size='large'), # CDRIVER-5874
    Distro(name='debian10-small', os='debian', os_type='linux', os_ver='10', size='small'), # CDRIVER-5874
    Distro(name='debian11-large', os='debian', os_type='linux', os_ver='11', size='large'),
    Distro(name='debian11-small', os='debian', os_type='linux', os_ver='11', size='small'),
    Distro(name='debian92-large', os='debian', os_type='linux', os_ver='9.2', size='large'),
    Distro(name='debian92-small', os='debian', os_type='linux', os_ver='9.2', size='small'),
]

MACOS_DISTROS = [
    Distro(name='macos-14', os='macos', os_type='macos', os_ver='14'),
]

MACOS_ARM64_DISTROS = [
    Distro(name='macos-11-arm64', os='macos', os_type='macos', os_ver='11', arch='arm64'),
    Distro(name='macos-14-arm64', os='macos', os_type='macos', os_ver='14', arch='arm64'),
]

RHEL_DISTROS = [
    Distro(name='rhel80-large', os='rhel', os_type='linux', os_ver='8.0', size='large'),
    Distro(name='rhel80-small', os='rhel', os_type='linux', os_ver='8.0', size='small'),
    Distro(name='rhel84-large', os='rhel', os_type='linux', os_ver='8.4', size='large'),
    Distro(name='rhel84-small', os='rhel', os_type='linux', os_ver='8.4', size='small'),
    Distro(name='rhel8.9-large', os='rhel', os_type='linux', os_ver='8.7', size='large'),
    Distro(name='rhel8.9-small', os='rhel', os_type='linux', os_ver='8.7', size='small'),
    Distro(name='rhel92-large', os='rhel', os_type='linux', os_ver='9.0', size='large'),
    Distro(name='rhel92-small', os='rhel', os_type='linux', os_ver='9.0', size='small'),
]

RHEL_ARM64_DISTROS = [
    Distro(name='rhel82-arm64-large', os='rhel', os_type='linux', os_ver='8.2', size='large', arch='arm64'),
    Distro(name='rhel82-arm64-small', os='rhel', os_type='linux', os_ver='8.2', size='small', arch='arm64'),
    Distro(name='rhel92-arm64-large', os='rhel', os_type='linux', os_ver='9.0', size='large', arch='arm64'),
    Distro(name='rhel92-arm64-small', os='rhel', os_type='linux', os_ver='9.0', size='small', arch='arm64'),
]

RHEL_POWER_DISTROS = [
    Distro(name='rhel8-power-large', os='rhel', os_type='linux', os_ver='8', size='large', arch='power'),
    Distro(name='rhel8-power-small', os='rhel', os_type='linux', os_ver='8', size='small', arch='power'),
    Distro(name='rhel9-power-large', os='rhel', os_type='linux', os_ver='9', size='large', arch='power'),
    Distro(name='rhel9-power-small', os='rhel', os_type='linux', os_ver='9', size='small', arch='power'),
]

RHEL_ZSERIES_DISTROS = [
    Distro(name='rhel8-zseries-large', os='rhel', os_type='linux', os_ver='8', size='large', arch='zseries'),
    Distro(name='rhel8-zseries-small', os='rhel', os_type='linux', os_ver='8', size='small', arch='zseries'),
    Distro(name='rhel9-zseries-large', os='rhel', os_type='linux', os_ver='9', size='large', arch='zseries'),
    Distro(name='rhel9-zseries-small', os='rhel', os_type='linux', os_ver='9', size='small', arch='zseries'),
]

UBUNTU_DISTROS = [
    Distro(name='ubuntu2004-large', os='ubuntu', os_type='linux', os_ver='20.04', size='large'),
    Distro(name='ubuntu2004-small', os='ubuntu', os_type='linux', os_ver='20.04', size='small'),
    Distro(name='ubuntu2204-large', os='ubuntu', os_type='linux', os_ver='22.04', size='large'),
    Distro(name='ubuntu2204-small', os='ubuntu', os_type='linux', os_ver='22.04', size='small'),
]

UBUNTU_ARM64_DISTROS = [
    Distro(name='ubuntu2004-arm64-large', os='ubuntu', os_type='linux', os_ver='20.04', size='large', arch='arm64'),
    Distro(name='ubuntu2004-arm64-small', os='ubuntu', os_type='linux', os_ver='20.04', size='small', arch='arm64'),
    Distro(name='ubuntu2204-arm64-large', os='ubuntu', os_type='linux', os_ver='22.04', size='large', arch='arm64'),
    Distro(name='ubuntu2204-arm64-small', os='ubuntu', os_type='linux', os_ver='22.04', size='small', arch='arm64'),
]

WINDOWS_DISTROS = [
    Distro(name='windows-vsCurrent-large', os='windows', os_type='windows', vs_ver='vsCurrent', size='large'), # Windows Server 2019
    Distro(name='windows-vsCurrent-small', os='windows', os_type='windows', vs_ver='vsCurrent', size='small'), # Windows Server 2019
]
#fmt: on
# pylint: enable=line-too-long

# Ensure no-arch distros are ordered before arch-specific distros.
ALL_DISTROS = [] + \
    DEBIAN_DISTROS + \
    MACOS_DISTROS + \
    MACOS_ARM64_DISTROS + \
    RHEL_DISTROS + \
    RHEL_ARM64_DISTROS + \
    RHEL_POWER_DISTROS + \
    RHEL_ZSERIES_DISTROS + \
    UBUNTU_DISTROS + \
    UBUNTU_ARM64_DISTROS + \
    WINDOWS_DISTROS


def find_distro(name) -> Distro:
    candidates = [d for d in ALL_DISTROS if name == d.name]

    if not candidates:
        raise ValueError(f'could not find a distro with the name {name}')

    return candidates[0]


def find_large_distro(name) -> Distro:
    candidates = [d for d in ALL_DISTROS if f'{name}-large' == d.name]

    if candidates:
        return candidates[0]

    return find_distro(name)


def find_small_distro(name) -> Distro:
    candidates = [d for d in ALL_DISTROS if f'{name}-small' == d.name]

    if candidates:
        return candidates[0]

    return find_distro(name)


def make_distro_str(distro_name, compiler, arch) -> str:
    if distro_name.startswith('windows-vsCurrent'):
        # Rename `windows-vsCurrent-*` distros to `windows-<year>` where`<year>`
        # is the Windows Server version used by the distro, e.g.:
        #     ('windows-vsCurrent-2022', 'vs2017x64', None) -> windows-2022-vs2017-x64
        #     ('windows-vsCurrent-2022', 'mingw',     None) -> windows-2022-mingw
        #     ('windows-vsCurrent',      'vs2017x64', None) -> windows-2019-vs2017-x64
        #     ('windows-vsCurrent',      'mingw',     None) -> windows-2019-mingw
        maybe_arch = compiler[len('vs20XY'):]
        if maybe_arch in ('x86', 'x64'):
            compiler_str = compiler[:-len(maybe_arch)] + '-' + maybe_arch
        else:
            compiler_str = compiler
        if distro_name.startswith('windows-vsCurrent-'):
            distro_str = 'windows-' + \
                distro_name[len('windows-vsCurrent-'):] + f'-{compiler_str}'
        else:
            distro_str = 'windows-2019' + f'-{compiler_str}'
    elif distro_name.startswith('windows-64-vs'):
        # Abbreviate 'windows-64-vs<type>' as 'vs<type>' and append '-<arch>' if
        # given in compiler string as 'vs<type><arch>', e.g.:
        #     ('windows-64-vs2017', 'vs2017x64', None) -> vs2017-x64
        #     ('windows-64-vs2017', 'mingw',     None) -> vs2017-mingw
        distro_str = distro_name[len('windows-64-'):] + {
            'vs2013x64': '-x64',
            'vs2013x86': '-x86',
            'vs2015x64': '-x64',
            'vs2015x86': '-x86',
            'vs2017x64': '-x64',
            'vs2017x86': '-x86',
        }.get(compiler, f'-{compiler}')
    else:
        distro_str = distro_name
        if compiler:
            distro_str += f'-{compiler}'

    if arch:
        distro_str += f'-{arch}'

    return distro_str


def to_cc(compiler):
    return {
        'vs2013x64': 'Visual Studio 12 2013 Win64',
        'vs2013x86': 'Visual Studio 12 2013',
        'vs2015x64': 'Visual Studio 14 2015 Win64',
        'vs2015x86': 'Visual Studio 14 2015',
        'vs2017x64': 'Visual Studio 15 2017 Win64',
        'vs2017x86': 'Visual Studio 15 2017',
    }.get(compiler, compiler)



"""
Download and extract MongoDB components.

Use '--help' for more information.
"""
import argparse
import enum
import hashlib
import json
import os
import platform
import re
import shutil
import sqlite3
import sys
import tarfile
import tempfile
import textwrap
import urllib.error
import urllib.request
import zipfile
from collections import namedtuple
from contextlib import contextmanager
from fnmatch import fnmatch
from pathlib import Path, PurePath, PurePosixPath

DISTRO_ID_MAP = {
    'elementary': 'ubuntu',
    'fedora': 'rhel',
    'centos': 'rhel',
    'mint': 'ubuntu',
    'linuxmint': 'ubuntu',
    'opensuse-leap': 'sles',
    'opensuse': 'sles',
    'redhat': 'rhel',
    'rocky': 'rhel',
    'ol': 'rhel',
}

DISTRO_VERSION_MAP = {
    'elementary': {
        '6': '20.04'
    },
    'fedora': {
        '32': '8',
        '33': '8',
        '34': '8',
        '35': '8',
        '36': '8'
    },
    'linuxmint': {
        '19': '18.04',
        '19.1': '18.04',
        '19.2': '18.04',
        '19.3': '18.04',
        '20': '20.04',
        '20.1': '20.04',
        '20.2': '20.04',
        '20.3': '20.04',
        '21': '22.04'
    },
}

DISTRO_ID_TO_TARGET = {
    'ubuntu': {
        '22.*': 'ubuntu2204',
        '20.*': 'ubuntu2004',
        '18.*': 'ubuntu1804',
        '16.*': 'ubuntu1604',
        '14.*': 'ubuntu1404',
    },
    'debian': {
        '9': 'debian92',
        '10': 'debian10',
        '11': 'debian11',
    },
    'rhel': {
        '6': 'rhel60',
        '6.*': 'rhel60',
        '7': 'rhel70',
        '7.*': 'rhel70',
        '8': 'rhel80',
        '8.*': 'rhel80',
    },
    'sles': {
        '10.*': 'suse10',
        '11.*': 'suse11',
        '12.*': 'suse12',
        '13.*': 'suse13',
        '15.*': 'suse15',
    },
    'amzn': {
        '2018.*': 'amzn64',
        '2': 'amazon2',
    },
}


def infer_target():
    if sys.platform == 'win32':
        return 'windows'
    if sys.platform == 'darwin':
        return 'macos'
    # Now the tricky bit
    cands = map(Path, ['/etc/os-release', '/usr/lib/os-release'])
    for c in cands:
        if c.is_file():
            return _infer_target_os_rel(c)
    raise RuntimeError("We don't know how to find the default '--target'"
                       " option for this system. Please contribute!")


def _infer_target_os_rel(os_rel_path: Path):
    with Path(os_rel_path).open('r', encoding='utf-8') as f:
        content = f.read()
    id_re = re.compile(r'\bID=("?)(.*)\1')
    mat = id_re.search(content)
    assert mat, 'Unable to detect ID from [{}] content:\n{}'.format(
        os_rel_path, content)
    os_id = mat.group(2)
    if os_id == 'arch':
        # There are no Archlinux-specific MongoDB downloads, so we'll just use
        # the build for RHEL8, which is reasonably compatible with other modern
        # distributions (including Arch).
        return 'rhel80'
    ver_id_re = re.compile(r'VERSION_ID=("?)(.*)\1')
    mat = ver_id_re.search(content)
    assert mat, 'Unable to detect VERSION_ID from [{}] content:\n{}'.format(
        os_rel_path, content)
    ver_id = mat.group(2)
    mapped_id = DISTRO_ID_MAP.get(os_id)
    if mapped_id:
        print('Mapping distro "{}" to "{}"'.format(os_id, mapped_id))
        ver_mapper = DISTRO_VERSION_MAP.get(os_id)
        if ver_mapper:
            mapped_version = ver_mapper[ver_id]
            print('Mapping version "{}" to "{}"'.format(
                ver_id, mapped_version))
            ver_id = mapped_version
        os_id = mapped_id
    os_id = os_id.lower()
    if os_id not in DISTRO_ID_TO_TARGET:
        raise RuntimeError("We don't know how to map '{}' to a distribution "
                           "download target. Please contribute!".format(os_id))
    ver_table = DISTRO_ID_TO_TARGET[os_id]
    for pattern, target in ver_table.items():
        if fnmatch(ver_id, pattern):
            return target
    raise RuntimeError(
        "We don't know how to map '{}' version '{}' to a distribution "
        "download target. Please contribute!".format(os_id, ver_id))


def caches_root():
    if sys.platform == 'win32':
        return Path(os.environ['LocalAppData'])
    if sys.platform == 'darwin':
        return Path(os.environ['HOME'] + '/Library/Caches')
    xdg_cache = os.getenv('XDG_CACHE_HOME')
    if xdg_cache:
        return Path(xdg_cache)
    return Path(os.environ['HOME'] + '/.cache')


def cache_dir():
    return caches_root().joinpath('mongodl').absolute()


@contextmanager
def tmp_dir():
    tdir = tempfile.mkdtemp()
    try:
        yield Path(tdir)
    finally:
        shutil.rmtree(tdir)


def _import_json_data(db, json_file):
    db.execute('DELETE FROM meta')
    db.execute('DROP TABLE IF EXISTS components')
    db.execute('DROP TABLE IF EXISTS downloads')
    db.execute('DROP TABLE IF EXISTS versions')
    db.execute(r'''
        CREATE TABLE versions (
            version_id INTEGER PRIMARY KEY,
            date TEXT NOT NULL,
            version TEXT NOT NULL,
            githash TEXT NOT NULL
        )
    ''')
    db.execute(r'''
        CREATE TABLE downloads (
            download_id INTEGER PRIMARY KEY,
            version_id INTEGER NOT NULL REFERENCES versions,
            target TEXT NOT NULL,
            arch TEXT NOT NULL,
            edition TEXT NOT NULL,
            ar_url TEST NOT NULL,
            data TEXT NOT NULL
        )
    ''')
    db.execute(r'''
        CREATE TABLE components (
            component_id INTEGER PRIMARY KEY,
            key TEXT NOT NULL,
            download_id INTEGER NOT NULL REFERENCES downloads,
            data TEXT NOT NULL,
            UNIQUE(key, download_id)
        )
    ''')
    with json_file.open('r') as f:
        data = json.load(f)
    for ver in data['versions']:
        version = ver['version']
        githash = ver['githash']
        date = ver['date']
        db.execute(
            r'''
            INSERT INTO versions (date, version, githash)
            VALUES (?, ?, ?)
            ''',
            (date, version, githash),
        )
        version_id = db.lastrowid
        for dl in ver['downloads']:
            arch = dl.get('arch', 'null')
            target = dl.get('target', 'null')
            edition = dl['edition']
            ar_url = dl['archive']['url']
            db.execute(
                r'''
                INSERT INTO downloads (version_id, target, arch, edition, ar_url, data)
                VALUES (?, ?, ?, ?, ?, ?)
                ''',
                (version_id, target, arch, edition, ar_url, json.dumps(dl)),
            )
            dl_id = db.lastrowid
            for key, data in dl.items():
                if 'url' not in data:
                    continue
                db.execute(
                    r'''
                    INSERT INTO components (key, download_id, data)
                    VALUES (?, ?, ?)
                    ''',
                    (key, dl_id, json.dumps(data)),
                )


def _mkdir(dirpath):
    """
    Ensure a directory at ``dirpath``, and all parent directories thereof.

    Cannot using Path.mkdir(parents, exist_ok) on some Python versions that
    we need to support.
    """
    if dirpath.is_dir():
        return
    par = dirpath.parent
    if par != dirpath:
        _mkdir(par)
    try:
        dirpath.mkdir()
    except FileExistsError:
        pass


def get_dl_db():
    caches = cache_dir()
    _mkdir(caches)
    db = sqlite3.connect(str(caches / 'downloads.db'), isolation_level=None)
    db.executescript(r'''
        CREATE TABLE IF NOT EXISTS meta (
            etag TEXT,
            last_modified TEXT
        )
    ''')
    db.executescript(r'''
        CREATE TABLE IF NOT EXISTS past_downloads (
            url TEXT NOT NULL UNIQUE,
            etag TEXT,
            last_modified TEXT
        )
    ''')
    changed, full_json = _download_file(
        db, 'https://downloads.mongodb.org/full.json')
    if not changed:
        return db
    with db:
        print('Refreshing downloads manifest ...')
        cur = db.cursor()
        cur.execute("begin")
        _import_json_data(cur, full_json)
    return db


def _print_list(db, version, target, arch, edition, component):
    if version or target or arch or edition or component:
        matching = db.execute(
            r'''
            SELECT version, target, arch, edition, key, components.data
              FROM components,
                   downloads USING(download_id),
                   versions USING(version_id)
            WHERE (:component IS NULL OR key=:component)
              AND (:target IS NULL OR target=:target)
              AND (:arch IS NULL OR arch=:arch)
              AND (:edition IS NULL OR edition=:edition)
              AND (:version IS NULL OR version=:version)
            ''',
            dict(version=version,
                 target=target,
                 arch=arch,
                 edition=edition,
                 component=component),
        )
        for version, target, arch, edition, comp_key, comp_data in matching:
            print('Download: {}\n\n'
                  ' Version: {}\n\n'
                  '  Target: {}\n\n'
                  '    Arch: {}\n\n'
                  ' Edition: {}\n\n'
                  '    Info: {}\n\n'.format(comp_key, version, target, arch,
                                            edition, comp_data))
        print('(Omit filter arguments for a list of available filters)')
        return

    arches, targets, editions, versions, components = next(
        iter(
            db.execute(r'''
        VALUES(
            (select group_concat(arch, ', ') from (select distinct arch from downloads)),
            (select group_concat(target, ', ') from (select distinct target from downloads)),
            (select group_concat(edition, ', ') from (select distinct edition from downloads)),
            (select group_concat(version, ', ') from (select distinct version from versions)),
            (select group_concat(key, ', ') from (select distinct key from components))
        )
        ''')))
    versions = '\n'.join(
        textwrap.wrap(versions,
                      width=78,
                      initial_indent='  ',
                      subsequent_indent='  '))
    targets = '\n'.join(
        textwrap.wrap(targets,
                      width=78,
                      initial_indent='  ',
                      subsequent_indent='  '))
    print('Architectures:\n'
          '  {}\n'
          'Targets:\n'
          '{}\n'
          'Editions:\n'
          '  {}\n'
          'Versions:\n'
          '{}\n'
          'Components:\n'
          '  {}\n'.format(arches, targets, editions, versions, components))


def infer_arch():
    a = platform.machine() or platform.processor()
    # Remap platform names to the names used for downloads
    return {
        'AMD64': 'x86_64',
    }.get(a, a)


DLRes = namedtuple('DLRes', ['is_changed', 'path'])


def _download_file(db, url):
    caches = cache_dir()
    info = list(
        db.execute(
            'SELECT etag, last_modified FROM past_downloads WHERE url=?',
            [url]))
    etag = None
    modtime = None
    if info:
        etag, modtime = info[0]
    headers = {}
    if etag:
        headers['If-None-Match'] = etag
    if modtime:
        headers['If-Modified-Since'] = modtime
    req = urllib.request.Request(url, headers=headers)
    digest = hashlib.md5(url.encode("utf-8")).hexdigest()[:4]
    dest = caches / 'files' / digest / PurePosixPath(url).name
    try:
        resp = urllib.request.urlopen(req)
    except urllib.error.HTTPError as e:
        if e.code != 304:
            raise
        return DLRes(False, dest)
    else:
        print('Downloading [{}] ...'.format(url))
        _mkdir(dest.parent)
        got_etag = resp.getheader("ETag")
        got_modtime = resp.getheader('Last-Modified')
        with dest.open('wb') as of:
            buf = resp.read(1024 * 1024 * 4)
            while buf:
                of.write(buf)
                buf = resp.read(1024 * 1024 * 4)
        db.execute(
            'INSERT OR REPLACE INTO past_downloads (url, etag, last_modified) VALUES (?, ?, ?)',
            (url, got_etag, got_modtime))
    return DLRes(True, dest)


def _dl_component(db, out_dir, version, target, arch, edition, component,
                  pattern, strip_components, test):
    print('Download {} v{}-{} for {}-{}'.format(component, version, edition,
                                                target, arch))
    matching = db.execute(
        r'''
        SELECT components.data
        FROM
            components,
            downloads USING(download_id),
            versions USING(version_id)
        WHERE
            target=:target
            AND arch=:arch
            AND edition=:edition
            AND version=:version
            AND key=:component
        ''',
        dict(version=version,
             target=target,
             arch=arch,
             edition=edition,
             component=component),
    )
    found = list(matching)
    if not found:
        raise ValueError(
            'No download for "{}" was found for '
            'the requested version+target+architecture+edition'.format(
                component))
    data = json.loads(found[0][0])
    cached = _download_file(db, data['url']).path
    return _expand_archive(cached,
                           out_dir,
                           pattern,
                           strip_components,
                           test=test)


def pathjoin(items):
    """
    Return a path formed by joining the given path components
    """
    return PurePath('/'.join(items))


def _test_pattern(path, pattern):
    """
    Test whether the given 'path' string matches the globbing pattern 'pattern'.

    Supports the '**' pattern to match any number of intermediate directories.
    """
    if pattern is None:
        return True
    # Convert to path objects
    path = PurePath(path)
    pattern = PurePath(pattern)
    # Split pattern into parts
    pattern_parts = pattern.parts
    if not pattern_parts:
        # An empty pattern always matches
        return True
    path_parts = path.parts
    if not path_parts:
        # Non-empty pattern requires more path components
        return False
    pattern_head = pattern_parts[0]
    pattern_tail = pathjoin(pattern_parts[1:])
    if pattern_head == '**':
        # Special "**" pattern matches and suffix of the path
        # Generate each suffix:
        tails = (path_parts[i:] for i in range(len(path_parts)))
        # Test if any of the suffixes match the remainder of the pattern:
        return any(_test_pattern(pathjoin(t), pattern_tail) for t in tails)
    if not fnmatch(path.parts[0], pattern_head):
        # Leading path component cannot match
        return False
    # The first component matches. Test the remainder:
    return _test_pattern(pathjoin(path_parts[1:]), pattern_tail)


class ExpandResult(enum.Enum):
    Empty = 0
    Okay = 1


def _expand_archive(ar, dest, pattern, strip_components, test):
    '''
    Expand the archive members from 'ar' into 'dest'. If 'pattern' is not-None,
    only extracts members that match the pattern.
    '''
    print('Extract from: [{}]'.format(ar.name))
    print('        into: [{}]'.format(dest))
    if ar.suffix == '.zip':
        n_extracted = _expand_zip(ar,
                                  dest,
                                  pattern,
                                  strip_components,
                                  test=test)
    elif ar.suffix == '.tgz':
        n_extracted = _expand_tgz(ar,
                                  dest,
                                  pattern,
                                  strip_components,
                                  test=test)
    else:
        raise RuntimeError('Unknown archive file extension: ' + ar.suffix)
    verb = 'would be' if test else 'were'
    if n_extracted == 0:
        if pattern and strip_components:
            print('NOTE: No files {verb} extracted. Likely all files {verb} '
                  'excluded by "--only={p}" and/or "--strip-components={s}"'.
                  format(p=pattern, s=strip_components, verb=verb))
        elif pattern:
            print('NOTE: No files {verb} extracted. Likely all files {verb} '
                  'excluded by the "--only={p}" filter'.format(p=pattern,
                                                               verb=verb))
        elif strip_components:
            print('NOTE: No files {verb} extracted. Likely all files {verb} '
                  'excluded by "--strip-components={s}"'.format(
                      s=strip_components, verb=verb))
        else:
            print('NOTE: No files {verb} extracted. Empty archive?'.format(
                verb=verb))
        return ExpandResult.Empty
    elif n_extracted == 1:
        print('One file {v} extracted'.format(v='would be' if test else 'was'))
        return ExpandResult.Okay
    else:
        print('{n} files {verb} extracted'.format(n=n_extracted, verb=verb))
        return ExpandResult.Okay


def _expand_tgz(ar, dest, pattern, strip_components, test):
    'Expand a tar.gz archive'
    n_extracted = 0
    with tarfile.open(str(ar), 'r:*') as tf:
        for mem in tf.getmembers():
            n_extracted += _maybe_extract_member(
                dest,
                mem.name,
                pattern,
                strip_components,
                mem.isdir(),
                lambda: tf.extractfile(mem),
                mem.mode,
                test=test,
            )
    return n_extracted


def _expand_zip(ar, dest, pattern, strip_components, test):
    'Expand a .zip archive.'
    n_extracted = 0
    with zipfile.ZipFile(ar, 'r') as zf:
        for item in zf.infolist():
            n_extracted += _maybe_extract_member(
                dest,
                item.filename,
                pattern,
                strip_components,
                item.is_dir(),
                lambda: zf.open(item, 'r'),
                0o655,
                test=test,
            )
    return n_extracted


def _maybe_extract_member(out, relpath, pattern, strip, is_dir, opener,
                          modebits, test):
    """
    Try to extract an archive member according to the given arguments.

    :return: Zero if the file was excluded by filters, one otherwise.
    """
    relpath = PurePath(relpath)
    print('  | {:-<65} |'.format(str(relpath) + ' '), end='')
    if len(relpath.parts) <= strip:
        # Not enough path components
        print(' (Excluded by --strip-components)')
        return 0
    if not _test_pattern(relpath, pattern):
        # Doesn't match our pattern
        print(' (excluded by pattern)')
        return 0
    stripped = pathjoin(relpath.parts[strip:])
    dest = Path(out) / stripped
    print('\n    -> [{}]'.format(dest))
    if test:
        # We are running in test-only mode: Do not do anything
        return 1
    if is_dir:
        _mkdir(dest)
        return 1
    with opener() as infile:
        _mkdir(dest.parent)
        with dest.open('wb') as outfile:
            shutil.copyfileobj(infile, outfile)
        os.chmod(str(dest), modebits)
    return 1


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    grp = parser.add_argument_group('List arguments')
    grp.add_argument('--list',
                     action='store_true',
                     help='List available components, targets, editions, and '
                     'architectures. Download arguments will act as filters.')
    dl_grp = parser.add_argument_group(
        'Download arguments',
        description='Select what to download and extract. '
        'Non-required arguments will be inferred '
        'based on the host system.')
    dl_grp.add_argument('--target',
                        '-T',
                        help='The target platform for which to download. '
                        'Use "--list" to list available targets.')
    dl_grp.add_argument('--arch',
                        '-A',
                        help='The architecture for which to download')
    dl_grp.add_argument(
        '--edition',
        '-E',
        help='The edition of the product to download (Default is "enterprise"). '
        'Use "--list" to list available editions.')
    dl_grp.add_argument(
        '--out',
        '-o',
        help='The directory in which to download components. (Required)',
        type=Path)
    dl_grp.add_argument('--version',
                        '-V',
                        help='The product version to download (Required). '
                        'Use "--list" to list available versions.')
    dl_grp.add_argument('--component',
                        '-C',
                        help='The component to download (Required). '
                        'Use "--list" to list available components.')
    dl_grp.add_argument(
        '--only',
        help=
        'Restrict extraction to items that match the given globbing expression. '
        'The full archive member path is matched, so a pattern like "*.exe" '
        'will only match "*.exe" at the top level of the archive. To match '
        'recursively, use the "**" pattern to match any number of '
        'intermediate directories.')
    dl_grp.add_argument(
        '--strip-path-components',
        '-p',
        dest='strip_components',
        metavar='N',
        default=0,
        type=int,
        help=
        'Strip the given number of path components from archive members before '
        'extracting into the destination. The relative path of the archive '
        'member will be used to form the destination path. For example, a '
        'member named [bin/mongod.exe] will be extracted to [<out>/bin/mongod.exe]. '
        'Using --strip-components=1 will remove the first path component, extracting '
        'such an item to [<out>/mongod.exe]. If the path has fewer than N components, '
        'that archive member will be ignored.')
    dl_grp.add_argument(
        '--test',
        action='store_true',
        help='Do not extract or place any files/directories. '
        'Only print what will be extracted without placing any files.')
    dl_grp.add_argument('--empty-is-error',
                        action='store_true',
                        help='If all files are excluded by other filters, '
                        'treat that situation as an error and exit non-zero.')
    args = parser.parse_args()
    db = get_dl_db()

    if args.list:
        _print_list(db, args.version, args.target, args.arch, args.edition,
                    args.component)
        return

    if args.version is None:
        raise argparse.ArgumentError(None, 'A "--version" is required')
    if args.component is None:
        raise argparse.ArgumentError(
            None, 'A "--component" name should be provided')
    if args.out is None:
        raise argparse.ArgumentError(None,
                                     'A "--out" directory should be provided')

    target = args.target or infer_target()
    arch = args.arch or infer_arch()
    edition = args.edition or 'enterprise'
    out = args.out or Path.cwd()
    out = out.absolute()
    result = _dl_component(db,
                           out,
                           version=args.version,
                           target=target,
                           arch=arch,
                           edition=edition,
                           component=args.component,
                           pattern=args.only,
                           strip_components=args.strip_components,
                           test=args.test)
    if result is ExpandResult.Empty:
        return 1
    return 0


if __name__ == '__main__':
    sys.exit(main())
